# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
from __future__ import absolute_import, division, print_function
# TODO: Python 3 compatibility

__all__ = ['Tracer', 'CubeShape', 'ShotgunRegistry', 'make_shotgun_data_z']

import os
import time
import sys
from subprocess import check_call
import numpy as np
import tqdm
import logging
import scipy.ndimage as ndimage

from elektronn2 import utils
from elektronn2.utils.utils_basic import unique_rows
from elektronn2.data import knossos_array, transformations
from elektronn2.data.skeleton import Trace

logger = logging.getLogger('elektronn2log')

if sys.version_info[:2] != (2, 7):
    raise ImportError(
        '\nSorry, this module only supports Python 2.7.'
        '\nYour current Python version is {}\n'.format(sys.version)
    )

try:
    from knossos_utils import skeleton as knossos_skeleton
    from knossos_utils.knossosdataset import KnossosDataset
except ImportError as e:
    logger.error('\nFor using the tracing_utils module, you will need to'
                 ' install the knossos_utils module'
                 ' (https://github.com/knossos-project/knossos_utils)\n')
    raise e


with open(os.devnull, 'w') as devnull:
    # mayavi is to dumb to raise an exception and instead crashes whole script....
    try:
        # "xset q" will always succeed to run if an X server is currently running
        check_call(['xset', 'q'], stdout=devnull, stderr=devnull)
        import mayavi.mlab as mlab
        from tvtk.util.ctf import ColorTransferFunction, PiecewiseFunction
        # Don't set backend explicitly, use system default...
    # if "xset q" fails, conclude that X is not running
    except: # (OSError, ImportError, CalledProcessError, ValueError)
        logger.warning("No mayavi imported, cannot plot skeletons")
        mlab = None


"""
Modus
m: memory feedback
r: radius scaling
k: maximum kernel interpolation
s: restrict scaling to >0.5

"""

class Tracer(object):
    def __init__(self, model, z_shift=0, data_source=None, bounding_box_zyx=None,
                 trace_kwargs=dict(aniso_scale=2), modus='m', shotgun_registry=None,
                 registry_interval=None, reference_radius=18.0):
        if isinstance(data_source, (knossos_array.KnossosArray, knossos_array.KnossosArrayMulti)):
            assert model.input_node.shape['f']==data_source.shape[0]
        elif isinstance(data_source, np.ndarray):
            assert model.input_node.shape['f']==data_source.shape[0]
        else:
            raise ValueError("Wrong Data source")
        if isinstance(model, str):
            raise ValueError("must pass a model")
            #model = neuromancer.model.modelload(model)

        self.rnn = False
        self.modus = modus
        if 'scan' in model.nodes:
            self.rnn = True
            self.n_steps = model['scan'].n_steps
            self.mem_hid0 = model['mem_hid'].get_value()
            print("Initial State mean=%.3f, std=%.3f" %(self.mem_hid0.mean(), self.mem_hid0.std()))
            self.mem_trace0 = np.zeros((1, 3*self.n_steps), dtype=np.float32)

        self.model = model
        self.z_shift = z_shift
        self.patch_size = np.array(model.input_node.shape.spatial_shape)
        self.data = data_source
        self.trace_kwargs = trace_kwargs
        self.bounding_box_zyx = np.array(bounding_box_zyx)
        self.model.predict_ext() # compile function
        self.shotgun_registry = shotgun_registry
        self.registry_interval = registry_interval
        self.reference_radius = reference_radius

    @staticmethod
    def zeropad(a, length):
        n = len(a)
        if n < length:
            return np.pad(a, [[length-n,0],[0,0]], mode='constant')
        else:
            return a

    def get_scale_factor(self, radius, old_factor, scale_strenght):
        # if old was large (zoom in), radius is smaller
        radius_true = radius * 1.0/old_factor
        # Equilibrium Condition scale_factor=1 <==> radius = 20
        new_factor = self.reference_radius / radius_true
        hi = scale_strenght * 2.0 + (1-scale_strenght) * 1.0 ### Original 1.5
        if 's' in self.modus:
            lo = scale_strenght * 0.5 + (1-scale_strenght) * 1.0 ### Original 0.3
        else:
            lo = scale_strenght * 0.3 + (1 - scale_strenght) * 1.0  ### Original 0.3

        new_factor = np.clip(new_factor, lo, hi)
        change = new_factor / old_factor
        change = np.clip(change, 0.75, 1.25)
        new_factor = old_factor * change

        return new_factor

    def trace(self, position_l, direction_il, count, gamma=0, trace_xyz=None,
              linked_skel=None, check_for_lost_track=True,
              check_for_uturn=False, check_bb=True, profile=False,
              info_str=None, reject_obb_traces=False, initial_scale=None):
        """
        Although psoition_l is in zyx order, the returned trace_obj is in xyz order
        """
        tt = utils.Timer()

        if trace_xyz is None:
            trace_xyz = Trace(linked_skel=linked_skel, **self.trace_kwargs)
            if self.shotgun_registry:
                self.shotgun_registry.new_trace(trace_xyz)

            prediction_c_dummy = np.array([0, 0, 0])
            trace_xyz.append(position_l[::-1], coord_cnn=prediction_c_dummy)

        n_gamma = max(gamma, 1)
        gammas = np.linspace(0, 2*np.pi, n_gamma, endpoint=False)
        i = 0
        break_all = False
        position_samples = utils.AccumulationArray(3)
        prediction_samples = utils.AccumulationArray(3)
        scale = 1.0 if initial_scale is None else initial_scale
        if self.rnn:
            assert len(gammas)==1
            mem_hid = self.mem_hid0
            mem_trace = self.mem_trace0
            self.model.predict_ext()
            memory_track = []
        else:
            self.model.predict() # this compiles the function

        if 'k' in self.modus:
            last_ch_max_interp = True
        else:
            last_ch_max_interp = False

        pbar = tqdm.tqdm(total=count, smoothing=0.15, desc=info_str, leave=False)
        end_message = ""
        continue_tracing = True
        features = np.array([17,0,0,0,0,0,0])
        while i < count:
            for gamma in gammas:
                try:
                    image, _, M = transformations.get_tracing_slice(self.data,
                                  self.patch_size, position_l, z_shift=self.z_shift,
                                  aniso_factor=2, sample_aniso=True, gamma=gamma,
                                  scale_factor=scale, direction_iso=direction_il,
                                  last_ch_max_interp=last_ch_max_interp)

                    trafo = transformations.Transform(M, position_l=position_l,
                                                      aniso_factor=2)
                    tt.check("Slicing", silent=True)
                except (transformations.WarpingOOBError, ValueError):
                    end_message = "Stop: out of BOUNCE after %g pix at %s"\
                    %(trace_xyz.runlength, position_l.astype(np.int))

                    break_all = True
                    break

                if self.rnn:
                    if 'trace_join' in self.model.nodes:
                        pred, gru = self.model.predict_ext(mem_trace, image[None], mem_hid)
                    else:
                        pred, gru = self.model.predict_ext(image[None], mem_hid)
                    prediction_c = pred[0,:3]
                    features = pred[0,3:]
                    if 'r' in self.modus:
                        radius = features[0] # this is only the predicted radius (scaled)
                        features[0] /= scale # make radius true
                        scale = self.get_scale_factor(radius, scale, 1.0)

                else:
                    prediction_c = self.model.predict(image[None])[0]
                    features = None

                tt.check("GPU", silent=True)
                position_l, direction_il = trafo.cnn_pred2lab_position(prediction_c)
                position_samples.append(position_l)
                prediction_samples.append(prediction_c)

            if len(trace_xyz)-len(trace_xyz.features)==1 and self.rnn: # repeat last feature prediction
                trace_xyz.features.append(features) # for first entry

            if break_all:
                break

            position_zyx = position_samples.mean()
            prediction_c = prediction_samples.mean()
            trace_xyz.append(position_zyx[::-1], coord_cnn=prediction_c, features=features)
            prediction_samples.clear()
            position_samples.clear()

            if self.rnn:
                memory_track.append(gru)
                if 'm' in self.modus:
                    mem_hid = gru

            if check_for_lost_track and trace_xyz.lost_track:
                end_message = "Stop: LOST TRACK after %g pix"\
                      %trace_xyz.runlength
                break

            if check_for_uturn and trace_xyz.uturn_occurred:
                end_message = "Stop: U-TURN after %g pix"\
                      % trace_xyz.runlength
                break

            if check_bb:
                if self.bounding_box_zyx is not None:
                    if np.any(position_zyx<self.bounding_box_zyx[0]) or \
                       np.any(position_zyx>self.bounding_box_zyx[1]):
                        end_message = "Stop: out of BBOX after %g pix at %s"\
                        %(trace_xyz.runlength,position_l.astype(np.int))
                        break

            if self.shotgun_registry and (i+2)%self.registry_interval==0 and i > 1:
                continue_tracing = self.shotgun_registry.check(trace_xyz)
                if not continue_tracing:
                    end_message = "Stop: MERGE with (%i) after %g pix at %s"\
                    %(self.shotgun_registry.last_merge_count,
                      trace_xyz.runlength, position_l.astype(np.int))
                    break

            if info_str:
                pbar.set_description(desc=info_str + ' s=%.2f' % scale)
            else:
                pbar.set_description(desc='s=%.2f' % scale)
            pbar.update()
            i += 1
            tt.check("Data struct", silent=True)

        pbar.close()

        if self.shotgun_registry and "MERGE" not in end_message: # check if OOB etc.
            continue_tracing = self.shotgun_registry.check(trace_xyz)
            if not continue_tracing:
                end_message += "\n + MERGE with(%i)"%self.shotgun_registry.last_merge_count
            else:
                end_message += " No MERGE"

        if end_message != "":
            tqdm.tqdm.write("\n"+end_message)

        tt.check("Data struct", silent=True)
        if reject_obb_traces and ("BBOX" in end_message) or \
           reject_obb_traces and ("BOUNCE" in end_message):
            tqdm.tqdm.write("REJECTING Trace because out of BBOX or OOB")
            if profile:
                tt.summary(print_func=tqdm.tqdm.write)
            return None

        if profile:
            tt.summary(print_func=tqdm.tqdm.write)

        if self.rnn:
            assert len(trace_xyz.features)==len(trace_xyz)
            return trace_xyz, memory_track
        else:
            return trace_xyz

    @staticmethod
    def perturb_direciton(direc, azimuth, polar):
        M = transformations.get_rotmat_from_direc(direc)[:3,:3] # dot(M[, b) = [1,0,0]
        M_inv = np.linalg.inv(M)
        direc_new = np.array([np.cos(polar),
                              np.sin(polar)*np.sin(azimuth),
                              np.sin(polar)*np.cos(azimuth)])
        direc_new = np.dot(M_inv, direc_new)
        return direc_new

    @staticmethod
    def plot_vectors(cv, vectors, fig=None):
        if fig is None:
            fig = mlab.figure(bgcolor=(1.0, 0.8, 0.4), size=(600,400))

        vectors = np.array(vectors)
        zero = np.zeros(vectors.shape[0])
        mlab.quiver3d([0],[0],[0], cv[0:1], cv[1:2], cv[2:3], figure=fig,color=(0, 0.0, 0.0), scale_factor=3, line_width=5)
        mlab.quiver3d(zero, zero, zero, *vectors.T, figure=fig,color=(0, 0.6, 0.2), scale_factor=3, line_width=3)
        mlab.quiver3d(zero, zero, zero, vectors[:,0], vectors[:,1], vectors[:,2], figure=fig,color=(0, 0.6, 0.2), scale_factor=3, line_width=3)

        return fig


class BallGenerator(object):
    def __init__(self, aniso):
        self.aniso = aniso
    @staticmethod
    def _make_ball(diameter, aniso):
        r2 = np.linspace(-1, 1, num=diameter) ** 2
        dist2 = r2[:, None, None] + r2[:, None] + r2
        ball = (dist2 <= 1.0 + 1e-3).astype(np.bool)
        if aniso:
            ball = ball[::2, :, :]
        return ball

    @utils.cache()
    def __call__(self, diameter):
        return self._make_ball(diameter, self.aniso)

ball_generator = BallGenerator(True)
ball_generator_iso = BallGenerator(False)

class ShotgunRegistry(object):
    def __init__(self, seeds_zyx, registry_extent, directions=None, debug=False,
                 radius_discout=0.5, check_w=3, occupied_thresh=0.6,
                 candidate_max_rel=0.75, candidate_max_min_margin=1.5):
        self.seeds_zyx = seeds_zyx
        self.registry_extent = np.array(registry_extent)
        self.mask = np.full(registry_extent, -1, dtype=np.int32)
        self.traces = []
        self.edges = []
        self.curr_seed_num = -1 # wrt to original seeds
        self.curr_trace_num = -1
        self.directions = directions
        self.last_merge_count = 0

        self.occupied_thresh = occupied_thresh
        self.radius_discount=radius_discout
        self.check_w = check_w
        self.candidate_max_rel = candidate_max_rel
        self.candidate_max_min_margin = candidate_max_min_margin
        self.debug_mask = None

        if debug:
            self.debug_mask = np.full(registry_extent, -1, dtype=np.uint16)

    def new_trace(self, trace):
        self.traces.append(trace)
        self.curr_trace_num += 1
        assert (self.curr_trace_num+1)==len(self.traces)

    def check(self, trace):
        """
        Check if trace goes into masked volume. If so, find out to which trace
        tree this belongs and merge.
        Return False to stop tracing
        Mask seeds and volume mask by current trace's log

        W: window length to do check on
        """
        if len(trace)<(self.check_w):
            return True

        coords_xyz = trace.coords[-self.check_w:]
        positions_zyx = np.round(coords_xyz[:,::-1]).astype(np.int)
        positions_zyx = np.clip(positions_zyx, [0,0,0], self.registry_extent-1)
        z,y,x = positions_zyx[:,0], positions_zyx[:,1], positions_zyx[:,2]
        occupied = (self.mask[z,y,x]>=0).mean() > self.occupied_thresh

        continue_tracing = True
        if occupied:
            continue_tracing = False
            ids = self.mask[z,y,x]
            ids = ids[ids>=0]
            indices, counts = np.unique(ids, return_counts=True)
            self.last_merge_count = len(indices)
            i = indices[counts.argmax()]
            if len(indices)>1:
                pass #check if the different indices are at least connected?

            self.edges.append([self.curr_trace_num, i] + [1,1,1])
            self.edges.append([i, self.curr_trace_num] + [1,1,1])

            # indices, distance_track = self.find_nearest_trace(coords_xyz)
            # self.last_merge_count = len(indices)
            # for i, t in zip(indices, distance_track):
            #     self.edges.append([self.curr_trace_num, i] + list(t))
            #     self.edges.append([i, self.curr_trace_num] + list(t))

        return continue_tracing


    def find_nearest_trace(self, coords_xyz):
        """
        Find all other tracks that are at least as close as 1.5 minimal
        (relative!) distance. (compare to closest point of each track)
        """
        distance_track = np.zeros((len(self.traces[:-1]), 3))
        for i,tr in enumerate(self.traces[:-1]):
            distances, indices, coordinates = tr.kdt.get_knn(coords_xyz, k=1)
            radii = tr.features[:,0][indices]
            k = np.argmin(distances/radii)
            distance_track[i] = [distances[k], radii[k], distances[k]/radii[k]]

        min_dist = (distance_track[:,2]).min()
        mask = (distance_track[:,2] < min_dist*self.candidate_max_min_margin) | (distance_track[:,2] < self.candidate_max_rel)
        distance_track_r = distance_track[mask]
        indices_r = mask.nonzero()[0]
        distance_track_r = np.maximum(distance_track_r, 1e-4) #
        return indices_r, distance_track_r


    def update_mask(self, coords_xyz, radii, index=None):
        radii = radii * self.radius_discount
        diameter = (radii*2).astype(np.int)
        diameter += 1-np.mod(diameter, 4)
        diameter = np.maximum(1, diameter)
        for c_xyz, d  in zip(coords_xyz, diameter):
            x,y,z = np.round(c_xyz).astype(np.int)
            r = d // 2
            r2 = d // 4
            ball = ball_generator(d)
            try:
                self.mask[z-r2:z+r2+1,y-r:y+r+1, x-r:x+r+1][ball] = index
                if index is not None and self.debug_mask is not None:
                    self.debug_mask[z-r2:z+r2+1,
                                    y-r :y+r+1,
                                    x-r :x+r+1] = ball*index
            except (ValueError, IndexError):
                print(z,y,x)
                print(ball.shape)
                print(self.mask[z-r2:z+r2+1,
                              y-r :y+r+1,
                              x-r :x+r+1].shape)
                print("trying to insert ball OBB?")

                raise RuntimeError


    def get_next_seed(self):
       max_count = len(self.seeds_zyx)
       position_l = None
       direc_il   = None
       i = self.curr_seed_num + 1
       while True:
           if i >= max_count:
               logger.info("All Seeds are done")
               break
           if  self.mask[tuple(self.seeds_zyx[i])] < 0:
               position_l = self.seeds_zyx[i]
               self.curr_seed_num = i
               if self.directions is None:
                   direc_il = (self.extent/2-position_l)
                   direc_il /= np.linalg.norm(direc_il)
               else:
                   direc_il = self.directions[i]

               break

           i += 1

       return position_l, direc_il


    def plot_mask_vol(self, figure=None, adjust_tfs=False):
        img = np.repeat(self.mask,2,axis=0).astype(np.float32)
        img = ndimage.filters.gaussian_filter(img, (1.0,1.0,1.0))
        img = img[::4,::4,::4]
        if figure is None:
            figure = mlab.figure()
        vol = mlab.pipeline.scalar_field(img)
        mlab.pipeline.volume(vol, color=(1,1,1), vmin=0.0-1e-3, vmax=1.0+1e-3, figure=figure)
        figure.scene.background = (1.0, 0.8, 0.4)
        figure.scene.parallel_projection = True
        volume = figure.children[0].children[0].children[0]
        volume.volume_mapper.reference_count = 3
        volume.volume_mapper.progress = 1.0
        volume.volume_mapper.interactive_sample_distance = 0.5
        volume.volume_mapper.sample_distance = 0.5
        volume.volume_mapper.maximum_image_sample_distance = 2.0
        volume.volume_property.scalar_opacity_unit_distance = 0.5
        volume.volume_property.specular = 0.3
        volume.volume_property.ambient = 0.2

        figure.scene.render()
        return figure


class CubeShape(object):
    def __init__(self, shape, offset=None, center=None, input_excess=None,
                 bbox_reduction=None):
        shape = np.array(shape)
        assert np.all(shape%2==0)

        self.shape = shape
        if offset is not None:
            assert center is None
            self.offset = np.array(offset)
            self.center = self.offset + self.shape.astype(np.float)/2 - 0.5
        elif center is not None:
            assert offset is None
            self.center =  np.array(center)
            self.offset = (self.center - self.shape//2 + 0.5).astype(np.int)
        else:
            raise ValueError()

        self.input_excess = np.array(input_excess)
        self.bbox_reduction = np.array(bbox_reduction)

    def shrink_off_sh_cent(self, amount):
        amount = np.array(amount)
        off = self.offset - amount
        sh  = self.shape + 2*amount
        cent = off + sh.astype(np.float)/2 - 0.5
        return off, sh, cent

    def input_off_sh_cent(self, input_excess=None):
        if input_excess is None and self.input_excess is not None:
            return self.shrink_off_sh_cent(self.input_excess)
        elif input_excess is not None:
            return self.shrink_off_sh_cent(input_excess)
        else:
            raise ValueError()

    def bbox_off_sh_cent(self, bbox_reduction=None):
        if bbox_reduction is None and self.bbox_reduction is not None:
            return self.shrink_off_sh_cent(-self.bbox_reduction)
        elif bbox_reduction is not None:
            return self.shrink_off_sh_cent(-bbox_reduction)
        else:
            raise ValueError()

    def bbox_wrt_input(self):
        assert self.input_excess is not None and self.bbox_reduction is not None
        lower = self.input_excess + self.bbox_reduction
        upper = self.input_excess + self.shape - 1*self.bbox_reduction
        return np.array([lower, upper])


    def bbox_wrt_self(self):
        assert self.input_excess is not None and self.bbox_reduction is not None
        lower = self.bbox_reduction
        upper = self.shape - 1*self.bbox_reduction
        return np.array([lower, upper])


def sort_seeds(seeds, directions, seed_values, first_dist, border, bbox):
    # first_dist: seeds_values larger than this are sorted before the others
    # border: [dz, dy, dx], thickness of border partition of cube, seeds from
    # the border are sorted prior to other seeds

    # first filter seeds which are not in bounding box
    bbox_mask = np.logical_and(np.all(seeds > bbox[0], axis=1),
                               np.all(seeds < bbox[1],axis=1))

    seeds = seeds[bbox_mask]
    seed_values = seed_values[bbox_mask]

    if len(seeds)==0:
        return None

    border = np.array(border)
    border_mask = np.logical_and(np.all(seeds > bbox[0] + border, axis=1),
                                 np.all(seeds < bbox[1] - border, axis=1))  # inner cube

    sort_mask = np.argsort(seed_values)[::-1] # is ascending!
    try:
        split = (seed_values[sort_mask] < first_dist).nonzero()[0][0]
    except IndexError: # if all values are larger than first dist
        split = 0

    border_mask = border_mask[sort_mask]
    faces0 = sort_mask[:split][~border_mask[:split]]
    inner0 = sort_mask[:split][border_mask[:split]]
    faces1 = sort_mask[split:][~border_mask[split:]]
    inner1 = sort_mask[split:][border_mask[split:]]
    new_index = np.hstack((faces0, inner0, faces1, inner1))

    seeds_new = seeds[new_index]
    seed_vals_new = seed_values[new_index]
    directions_new = directions[new_index]
    return seeds_new, directions_new, seed_vals_new

@utils.timeit
def make_shotgun_data_z(cube_shape, save_name, z_skip=5):
    barr_thresh = 0.7
    z_lookaround = 5
    max_footprint_S = ball_generator(9)
    max_footprint_L = ball_generator(21)
    max_footprint_L = max_footprint_L[3:-3]

    peak_thresh = 3.0
    peak_thresh_L = 9

    kds_barr = KnossosDataset()
    data_prefix = os.path.expanduser("~/lustre/sdorkenw/j0126_")
    kds_barr.initialize_from_knossos_path(data_prefix+'161012_barrier/')
    pred = kds_barr.from_raw_cubes_to_matrix(cube_shape.shape,
                                             cube_shape.offset,
                                             show_progress=False,
                                             zyx_mode=True, datatype=np.float32)
    pred /= 255

    mem_high = np.invert(pred > barr_thresh)
    seeds = []
    seed_values = []
    noise = np.random.rand(*(mem_high[0:0 + z_lookaround].shape)) * 1e-3
    running_sum = 0
    for z in range(0, mem_high.shape[0] - z_lookaround, z_skip):
        try:
            dt = ndimage.distance_transform_edt(mem_high[z:z + z_lookaround], sampling=[2, 1, 1])
            dt = ndimage.filters.gaussian_filter(dt, (1.0,2.0,2.0))
            dt += noise

            z_peaks_S = ndimage.maximum_filter(dt, footprint=max_footprint_S,mode='constant')
            z_peaks_L = ndimage.maximum_filter(dt, footprint=max_footprint_L,mode='constant')

            z_peaks_small = (z_peaks_S == dt) * ((peak_thresh_L > dt) & (dt > peak_thresh))
            z_peaks_large = (z_peaks_L == dt) * ((peak_thresh_L <= dt))
            z_peaks = z_peaks_large + z_peaks_small
            z_peaks *= (pred[z:z + z_lookaround] < 0.5)
            seeds_z = np.array(z_peaks.nonzero()).T
            seeds_z[:, 0] += z
        except KeyboardInterrupt:
            break
        finally:
            seeds.append(seeds_z)
            seed_values.append(dt[z_peaks])
            running_sum += z_peaks.sum()
            print(z, running_sum, z_peaks_small.sum(), z_peaks_large.sum(), z_peaks.sum())

    seeds = np.concatenate(seeds, axis=0)
    seed_values = np.concatenate(seed_values, axis=0)
    seeds, index = unique_rows(seeds)
    seed_values = seed_values[index]


    lar = np.array([4,8,8])
    lari = lar * [2,1,1]
    sz  = lar * 2 + 1
    szi = lari * 2 + 1

    pred2 = kds_barr.from_raw_cubes_to_matrix(cube_shape.shape + 2 * lar,
                                              cube_shape.offset - lar,
                                              show_progress=False,
                                              zyx_mode=True, datatype=np.float32)
    pred2 /= 255

    mem_high2 = np.invert(pred2 > barr_thresh)
    dt = ndimage.distance_transform_edt(mem_high2, sampling=[2, 1, 1])
    local_grid = np.vstack([x-x.mean() for x in np.ones(szi).nonzero()])
    directions = np.zeros((len(seeds), 3))
    perm = np.random.permutation(local_grid.shape[1])[:400]
    for i, (s, v) in enumerate(zip(seeds, seed_values)):
        z, y, x = s  # np.round(s).astype(np.int)
        cavity = dt[z:z + sz[0], y:y + sz[1], x:x + sz[2]]
        cavity = ndimage.zoom(cavity, [float(sz[1])/sz[0],1,1])
        s_val = dt[z + lar[0], y + lar[1], x + lar[2]]
        diff = np.abs(cavity - s_val)
        d_m = diff.mean()
        mask = (diff < d_m)
        d_max = diff[mask].max()
        um = np.zeros_like(diff)
        um[mask] = d_max - diff[mask]
        um = um.ravel()
        uu, dd, vv = np.linalg.svd((um * local_grid).T[perm])
        direc_iso = vv[0]  # take largest eigenvector
        direc_iso /= np.linalg.norm(direc_iso, axis=0)  # normalise
        directions[i] = direc_iso

#    local_grid = np.mgrid[-2:2:5j, -2:2:5j,-2:2:5j]
#    local_grid = np.vstack([g.ravel() for g in local_grid])
#    directions = np.zeros((len(seeds), 3))
#    for i, s in enumerate(seeds):
#        z, y, x = s  # np.round(s).astype(np.int)
#        cavity = pred2[z:z + 3, y:y + 5, x:x + 5]
#        cavity = ndimage.zoom(cavity, [5.0/3.0,1,1] )
#        cavity = (1 - cavity).ravel()
#        uu, dd, vv = np.linalg.svd((cavity * local_grid).T)
#        direc_iso = vv[0]  # take largest eigenvector
#        direc_iso /= np.linalg.norm(direc_iso, axis=0)  # normalise
#        directions[i] = direc_iso

    seeds += cube_shape.offset
    utils.picklesave([seeds, directions, seed_values], save_name)

    # Creater skeleton with seeds and dirs
    skel_obj = knossos_skeleton.Skeleton()
    anno = knossos_skeleton.SkeletonAnnotation()
    anno.scaling = (9.0, 9.0, 20.0)
    skel_obj.add_annotation(anno)
    def add_node(s, r=5):
        z,y,x = s
        new_node = knossos_skeleton.SkeletonNode()
        new_node.from_scratch(anno, x, y, z, radius=r)
        anno.addNode(new_node)
        return new_node

    for s, dir, v in zip(seeds, directions, seed_values):
        n = add_node(s, r=4)
        n.appendComment("%.1f" %v)
        dir = dir.copy()
        dir[0] /= 2
        n1 = add_node((s + 11 * dir), r=1)
        n2 = add_node((s - 11 * dir), r=1)
        n1.addChild(n)
        n2.addChild(n)

    return seeds, directions, seed_values, skel_obj


def rebuild_model(model=None, path=None, save_name=None):
    from elektronn2 import neuromancer
    if model is None:
        model = neuromancer.model.modelload(path)

    model = neuromancer.model.rebuild_rnn(model)
    fail_prob = list(filter(lambda x: 'failing_prob' in x, model.nontrainable_params.keys()))
    if fail_prob:
        assert len(fail_prob)==1
        model.nontrainable_params[fail_prob[0]].set_value(0.0)
        print("NOTE: Setting fail prob to 0")

    if len(model.dropout_rates):
         model.dropout_rates = 0.0
         print("NOTE: Setting dropout_rates to 0")

    if save_name is not None:
        model.save(save_name)
    return model

def get_stitching_edges(traces_a, traces_b, k=2, rel_dist_thresh=1.2):
    edges = []
    #stats = []
    #stats_mean = []
    for i_a,t_a in enumerate(traces_a):
        for i_b,t_b in enumerate(traces_b):
            if len(t_a) < len(t_b):
                t_long = t_b
                t_short = t_a
            else:
                t_long = t_b
                t_short = t_a

            tmp = t_long.kdt.get_knn(t_short.coords.data, k=1)
            distances, indices, coords = tmp
            min_ix = distances.argsort()[:k]
            distances = distances[min_ix]
            indices = indices[min_ix]
            radii = (t_long.features[indices,0] + t_short.features[min_ix,0]) / 2
            rel_distances = distances / radii.mean()
            rel_mean = rel_distances.mean()
            #stats.extend(rel_distances)
            #stats_mean.append(rel_mean)

            if rel_mean < rel_dist_thresh:
                edges.append((i_a, i_b, rel_mean))

    #stats = np.array(stats)
    #stats_mean = np.array(stats_mean)
    return edges

def find_job(args):
    a, t_a, bs, trees_b, rel_dist_thresh, rel_search_radius, max_dist, thresh, k, dir_name = args
    print(a)
    edges = []
    for b, t_b in zip(bs, trees_b):
        if len(t_a.joined_coords) < len(t_b.joined_coords):
            t_long = t_b
            t_short = t_a
        else:
            t_long = t_b
            t_short = t_a

        tmp = t_long.joined_kdt.get_knn(t_short.joined_coords, k=1)
        distances, indices, coords = tmp

        min_ix = distances.argsort()[:k]
        distances = distances[min_ix]
        indices = indices[min_ix]
        radii = (t_long.joined_radii[indices] + t_short.joined_radii[min_ix]) / 2
        rel_distances = distances / radii.mean()
        rel_mean = rel_distances.mean()
        if rel_mean < rel_dist_thresh:
            pass #this should have been added already
        elif rel_mean < rel_search_radius and distances.mean() < max_dist:
            # Warning this works on isotropic coords
            closest_point_long = coords[min_ix[0]].copy()
            closest_point_short = t_short.joined_coords[min_ix[0]].copy()
            try:
                tmp = t_long.joined_kdt.get_knn(closest_point_short, k=20)
            except ValueError:
                k_ = len(t_long.joined_coords)
                tmp = t_long.joined_kdt.get_knn(closest_point_short, k=k_)

            coords_long = tmp[2].copy()
            try:
                tmp = t_short.joined_kdt.get_knn(closest_point_long, k=20)
            except ValueError:
                k_ = len(t_short.joined_coords)
                tmp = t_short.joined_kdt.get_knn(closest_point_long, k=k_)

            coords_short = tmp[2].copy()

            closest_point_long[2] *= 2
            closest_point_short[2] *= 2
            coords_long[:, 2] *= 2
            coords_short[:, 2] *= 2

            center_long = coords_long.mean(0)
            center_short = coords_short.mean(0)

            uu, dd, vv = np.linalg.svd(coords_long - center_long)
            direc_long = vv[0]  # take largest eigenvector

            uu, dd, vv = np.linalg.svd(coords_short - center_short)
            direc_short = vv[0]  # take largest eigenvector

            direc_centers = center_long - center_short
            direc_centers /= np.linalg.norm(direc_centers)

            direc_pairs = closest_point_long - closest_point_short
            direc_pairs /= np.linalg.norm(direc_pairs)

            # algin all dirctions:
            if np.dot(direc_short, direc_centers) < 0:
                direc_short *= -1
            if np.dot(direc_long, direc_centers) < 0:
                direc_long *= -1
            if np.dot(direc_pairs, direc_centers) < 0:
                direc_pairs *= -1

            m1 = np.dot(direc_short, direc_centers)
            m2 = np.dot(direc_long, direc_centers)
            m = (m1 + m2) / 2
            print(m1, m2, m)

            if m>thresh:
                edges.append((a,b))


            ###############################################################
            skel_obj = knossos_skeleton.Skeleton()
            anno = knossos_skeleton.SkeletonAnnotation()
            anno.scaling = (9.0, 9.0, 20.0)
            skel_obj.add_annotation(anno)
            anno.setComment("%.1f %.1f %.1f" %(m, m1, m2,))

            def add_node(x, y, z, r=5):
                new_node = knossos_skeleton.SkeletonNode()
                new_node.from_scratch(anno, x, y, z // 2, radius=r)
                anno.addNode(new_node)
                return new_node

            n0 = add_node(*center_long, r=10)
            n1 = add_node(*center_short, r=10)
            n0.addChild(n1)

            n2 = add_node(*closest_point_long, r=5)
            n3 = add_node(*closest_point_short, r=5)
            n2.addChild(n3)

            n4 = add_node(*(center_long + 20 * direc_long), r=3)
            n5 = add_node(*(center_long - 20 * direc_long), r=3)
            n0.addChild(n4)
            n0.addChild(n5)

            n6 = add_node(*(center_short + 20 * direc_short), r=3)
            n7 = add_node(*(center_short - 20 * direc_short), r=3)
            n1.addChild(n6)
            n1.addChild(n7)

            if not os.path.exists("merge_candidates"):
                os.makedirs("merge_candidates")

            skel_obj.to_kzip("%s/splitfix_candidates/%i.k.zip" %(dir_name, time.time()))
            ###############################################################

    return edges

def find_false_splits(simple_trees, k=2, thresh=0.85, rel_dist_thresh=1.2,
                        rel_search_radius=6, max_dist=40.0, dir_name=''):
    n = len(simple_trees)
    var_args = []
    for a  in range(n):
        t_a = simple_trees[a]
        bs = range(a + 1, n)
        trees_b = simple_trees[slice(a+1, n)]
        var_args.append([a, t_a, bs, trees_b])

    const_args = (rel_dist_thresh, rel_search_radius, max_dist, thresh, k, dir_name)
    edges = utils.parallel_accum(find_job, 1, var_args, const_args, proc=20)
    return np.array(edges)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
#
# The following code is copied (with only marginal changes) from
# https://github.com/ilastik/lazyflow/blob/f8b4658c9894fc1fd54f36dedcf5c0bf324fccc1/lazyflow/utility/fileLock.py
# See below for original license information.
# Thanks to Evan Fosmark, the ilastik team and other contributors to this code.
#
#
#
################################################################################
#   lazyflow: data flow based lazy parallel computation framework
#
#       Copyright (C) 2011-2014, the ilastik developers
#                                <team@ilastik.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the Lesser GNU General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# See the files LICENSE.lgpl2 and LICENSE.lgpl3 for full text of the
# GNU Lesser General Public License version 2.1 and 3 respectively.
# This information is also available on the ilastik web site at:
#		   http://ilastik.org/license/
###############################################################################
"""
Implementation of a simple cross-platform file locking mechanism.
This is a modified version of code retrieved on 2013-01-01 from 
http://www.evanfosmark.com/2009/01/cross-platform-file-locking-support-in-python.
(The original code was released under the BSD License.  See below for details.)

Modifications in this version:
 - Tweak docstrings for sphinx.
 - Accept an absolute path for the protected file (instead of a file name relative to cwd).
 - Allow timeout to be None.
 - Fixed a bug that caused the original code to be NON-threadsafe when the same FileLock instance was shared by multiple threads in one process.
   (The original was safe for multiple processes, but not multiple threads in a single process.  This version is safe for both cases.)
 - Added ``purge()`` function.
 - Added ``available()`` function.
 - Expanded API to mimic ``threading.Lock interface``:
   - ``__enter__`` always calls ``acquire()``, and therefore blocks if ``acquire()`` was called previously.
   - ``__exit__`` always calls ``release()``.  It is therefore a bug to call ``release()`` from within a context manager.
   - Added ``locked()`` function. 
   - Added blocking parameter to ``acquire()`` method

WARNINGS: 
 - The locking mechanism used here may need to be changed to support old NFS filesystems:
   http://lwn.net/Articles/251004
   (Newer versions of NFS should be okay, e.g. NFSv3 with Linux kernel 2.6. Check the open(2) man page for details about O_EXCL.)
 - This code has not been thoroughly tested on Windows, and there has been one report of incorrect results on Windows XP and Windows 7.
   The locking mechanism used in this class should (in theory) be cross-platform, but use at your own risk.

ORIGINAL LICENSE:

The original code did not properly include license text.
(It merely said "License: BSD".)
Therefore, we'll attach the following generic BSD License terms to this file.
Those who extract this file from the lazyflow code base (LGPL) for their own use
are therefore bound by the terms of both the Simplified BSD License below AND the LGPL.

Copyright (c) 2013, Evan Fosmark and others.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the FreeBSD Project.
"""

import os
import sys
import time
import errno


class FileLock(object):
    """ A file locking mechanism that has context-manager support so 
        you can use it in a ``with`` statement. This should be relatively cross
        compatible as it doesn't rely on ``msvcrt`` or ``fcntl`` for the locking.
    """

    class FileLockException(Exception):
        pass

    def __init__(self, protected_file_path, timeout=120, delay=1,
                 lock_file_contents=None):
        """ Prepare the file locker. Specify the file to lock and optionally
            the maximum timeout and the delay between each attempt to lock.
        """
        self.is_locked = False
        self.lockfile = protected_file_path + ".lock"
        self.timeout = timeout
        self.delay = delay
        self._lock_file_contents = lock_file_contents
        if self._lock_file_contents is None:
            self._lock_file_contents = "Owning process args:\n"
            for arg in sys.argv:
                self._lock_file_contents += arg + "\n"

    def locked(self):
        """
        Returns True iff the file is owned by THIS FileLock instance.
        (Even if this returns false, the file could be owned by another FileLock instance, possibly in a different thread or process).
        """
        return self.is_locked

    def available(self):
        """
        Returns True iff the file is currently available to be locked.
        """
        return not os.path.exists(self.lockfile)

    def acquire(self, blocking=True):
        """ Acquire the lock, if possible. If the lock is in use, and `blocking` is False, return False.
            Otherwise, check again every `self.delay` seconds until it either gets the lock or
            exceeds `timeout` number of seconds, in which case it raises an exception.
        """
        start_time = time.time()
        while True:
            try:
                # Attempt to create the lockfile.
                # These flags cause os.open to raise an OSError if the file already exists.
                fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                with os.fdopen(fd, 'a') as f:
                    # Print some info about the current process as debug info for anyone who bothers to look.
                    f.write(self._lock_file_contents)
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if self.timeout is not None and (
                    time.time() - start_time) >= self.timeout:
                    raise FileLock.FileLockException("Timeout occurred.")
                if not blocking:
                    return False
                time.sleep(self.delay)
        self.is_locked = True
        return True

    def release(self):
        """ Get rid of the lock by deleting the lockfile. 
            When working in a `with` statement, this gets automatically 
            called at the end.
        """
        self.is_locked = False
        os.unlink(self.lockfile)

    def __enter__(self):
        """ Activated when used in the with statement. 
            Should automatically acquire a lock to be used in the with block.
        """
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """ Activated at the end of the with statement.
            It automatically releases the lock if it isn't locked.
        """
        self.release()

    def __del__(self):
        """ Make sure this ``FileLock`` instance doesn't leave a .lock file
            lying around.
        """
        if self.is_locked:
            self.release()

    def purge(self):
        """
        For debug purposes only.  Removes the lock file from the hard disk.
        """
        if os.path.exists(self.lockfile):
            self.release()
            return True
        return False


if __name__ == "__main__":
    import sys
    import functools
    import threading
    import tempfile

    temp_dir = tempfile.mkdtemp()
    protected_filepath = os.path.join(temp_dir, "somefile.txt")
    "Protecting file: {}".format(protected_filepath)
    fl = FileLock(protected_filepath)


    def writeLines(line, repeat=10):
        with fl:
            for _ in range(repeat):
                with open(protected_filepath, 'a') as f:
                    f.write(line + "\n")
                    f.flush()


    th1 = threading.Thread(
        target=functools.partial(writeLines, "1111111111111111111111111111111"))
    th2 = threading.Thread(
        target=functools.partial(writeLines, "2222222222222222222222222222222"))
    th3 = threading.Thread(
        target=functools.partial(writeLines, "3333333333333333333333333333333"))
    th4 = threading.Thread(
        target=functools.partial(writeLines, "4444444444444444444444444444444"))

    th1.start()
    th2.start()
    th3.start()
    th4.start()

    th1.join()
    th2.join()
    th3.join()
    th4.join()

    assert not os.path.exists(fl.lockfile), "The lock file wasn't cleaned up!"

    # Print the contents of the file.
    # Please manually inspect the output.  Does it look like the operations were atomic?
    with open(protected_filepath, 'r') as f:
        sys.stdout.write(f.read())



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

import copy
import logging

import numpy as np
import theano.tensor as T
from theano.tensor.sharedvar import TensorSharedVariable
from theano.tensor import TensorConstant, TensorType

from .. import config
from .graphutils import floatX, as_floatX


logger = logging.getLogger('elektronn2log')

__all__ = ['VariableParam', 'VariableWeight', 'ConstantParam', 'initweights']

class VariableParam(TensorSharedVariable):
    """
    Extension of theano ``TensorSharedVariable``. Additional features are
    described by the parameters, otherwise identical

    Parameters
    ----------
    value
    name: str
    apply_reg flag: bool
        whether to apply regularisation (e.g. L2) on this param
    apply_train flag: bool
        whether to train this parameter (as opposed to a meta-parameter or
        a parameter that is kept const. during a training phase)
    dtype:
    strict: bool
    allow_downcast: bool
    borrow: bool
    broadcastable
    """

    def __init__(self, value=None, name=None, apply_train=True, apply_reg=True,
                 dtype=None, strict=False, allow_downcast=None, borrow=False,
                 broadcastable=None):
        self.apply_reg   = apply_reg
        self.apply_train = apply_train
        self._updates     = None
        self.constant    = False
        if not apply_train:
            name += "_noTrain"

        if isinstance(value, (int, float)):
            value = np.array(value, dtype=dtype)

        if dtype is not None:
            value = value.astype(dtype)

        value=np.array(value, copy=(not borrow))

        if broadcastable is None:
            broadcastable = (False,) * len(value.shape)
        t_type = T.TensorType(value.dtype, broadcastable=broadcastable)

        super(VariableParam, self).__init__(type=t_type, value=value, name=name,
                                          strict=strict,
                                          allow_downcast=allow_downcast)

    @property
    def updates(self):
        return self._updates

    @updates.setter
    def updates(self, up):
        if self.apply_train or self.apply_reg:
            raise ValueError("Cannot register extra updates for trainable "
                             "parameter %s" %(repr(self),))
        self._updates = up

    def clone(self):
        cp = TensorSharedVariable(
            name=self.name,
            type=self.type,
            value=None,
            strict=None,
            container=self.container)
        cp.tag = copy.copy(self.tag)
        return cp


class VariableWeight(VariableParam):
    def __init__(self, shape=None, init_kwargs=None, value=None, name=None,
                 apply_train=True, apply_reg=True, dtype=None, strict=False,
                 allow_downcast=None, borrow=False, broadcastable=None):
        """
        Extension of theano ``TensorSharedVariable`` and subclass of ``VariableParam``.
        Additional features are described by the parameters, otherwise identical

        Parameters
        ----------
        shape: list/tuple of int
            Shape of weights (if value=None)
        init_kwargs: dict
            kwargs for the ``initweights``-function (if value=None)
        value: numpy array
            initial value (if shape/init_kwargs=None)
        name: str
        apply_train flag: bool
            Whether to train this parameter (as opposed to a meta-parameter or
            a parameter that is kept const. during a training phase)
        apply_reg flag: bool
            Whether to apply regularisation (e.g. L2) on this param
        dtype
        strict: bool
        allow_downcast: bool
        borrow: bool
        broadcastable
        """
        if value is None: # create new values
            if (shape is None) or (init_kwargs is None):
                raise ValueError("shape and init_kwargs are required if value is None")
            value = initweights(shape, **init_kwargs)

        elif shape is not None:
            if np.array(value).ndim > 1:
                raise ValueError("If value and shape are specified, "
                                 "value must be scalar.")
            value = np.ones(shape) * value

        super(VariableWeight, self).__init__(value, name, apply_train, apply_reg,
                                           dtype, strict, allow_downcast,
                                           borrow, broadcastable)

    def set_value(self, new_value, borrow=False):
        sh = self.get_value().shape
        if isinstance(new_value, np.ndarray):
            if not (sh == new_value.shape):
                raise NotImplementedError("given shape: %s, required shape: %s "
                "Crop value or extend with similar numbers"%(new_value.shape, sh))
        elif isinstance(new_value, (float, int)):
            pass
        else:
            raise ValueError("Value/type not understood")

        try:
            super(VariableWeight, self).set_value(new_value, borrow)
        except TypeError as e:
            if config.allow_floatX_downcast:
                new_value = as_floatX(new_value)
                super(VariableWeight, self).set_value(new_value, borrow)
            else:
                raise

class ConstantParam(TensorConstant):
    """
    Identical to theano ``VariableParam`` except that there are two
    two addition attributes ``apply_train`` and `apply_reg``, which are
    both false.
    This is just to tell ELEKTRONN2 that this parameter is to be
    exempted from training. Obviously the ``set_value`` method raises
    an exception because this is a real constant. Constants are faster
    in the theano graph.
    """

    def __init__(self, value, name=None, dtype=None,
                 make_singletons_broadcastable=True):
        name += "_const"

        if isinstance(value, (int, float)):
            value = np.array(value, dtype=dtype)

        if dtype is not None:
            value = value.astype(dtype)
        if make_singletons_broadcastable:
            broadcastable = [d == 1 for d in value.shape]
        else:
            broadcastable = [False for d in value.shape]

        dtype_t = TensorType(dtype=value.dtype, broadcastable=broadcastable)

        self.apply_train = False
        self.apply_reg   = False
        self.constant    = True

        super(ConstantParam, self).__init__(dtype_t, value, name=name)

    def clone(self):
        return TensorConstant(self.type, self.data, self.name)

    def set_value(self, new_value, borrow=False):
        raise RuntimeError("Cannot set value for ConstantParam")


    def get_value(self, borrow=False):
        return self.value

    @property
    def updates(self):
        return None


def initweights(shape, dtype=floatX, scale='glorot', mode='normal', pool=None, spatial_axes=None):
    if mode=='const':
        W = np.ones(shape) * scale

    elif mode=='prelu':
        # assuming shape is (n_out, 2)
        W = np.ones(shape) * scale
        W[:,1] = 1.0
        # GradNet inspired initial full linearity, conventional relu would be 0

    elif mode=='fix-uni':
        W = np.random.uniform(-scale, scale, shape)

    elif scale=='glorot':
        if len(shape)==2: # (fin, nof)
            n_in, n_out = shape[0], shape[1]
            s = n_in + n_out
        else:
            assert spatial_axes is not None
            other, kernel = [], []
            for i,s in enumerate(shape):
                if i in spatial_axes:
                    kernel.append(s)
                else:
                    other.append(s)

            assert len(other)==2
            n_out = other[0]
            n_in  = other[1]
            fov = np.prod(kernel)
            ps  = np.prod(pool)
            s = (n_in + float(n_out)/ps) * fov

        W_scale = np.sqrt(2.0 / s)

        if mode=='normal':
            W = np.random.normal(0, W_scale, shape)
        elif mode=='uni':
            W = np.random.uniform(-W_scale, W_scale, shape)
        elif mode=='ortho':
            M = np.random.normal(0, W_scale, size=shape)
            M = M.reshape((n_out, -1))
            # more vectors needed than can be orthogonal in this dimension
            strip_required = False
            n_in = M.shape[1]
            if n_out > n_in:
                M = np.random.normal(0, W_scale, size=(n_out, n_out))
                strip_required = True

            U, S, V = np.linalg.svd(M, full_matrices=False)
            W = V / V.std(1)[:,None] * W_scale
            #W -= W.mean(axis=1)[:,None] # This changes whether they are orthogonal!
            if strip_required:
                W = W[:, :n_in]

            W = W.reshape(shape)
    else:
        raise ValueError("Invalid weigh initialisation parameters")

    logger.debug("Init: shape=%s, mean=%f, std=%f"%(shape, W.mean(), W.std()))

    return np.ascontiguousarray(W, dtype=dtype)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


__all__ = ['make_affinities', 'downsample_xy', 'ids2barriers', 'smearbarriers',
           'center_cubes', ]


import logging
import multiprocessing
from functools import reduce

import numba
from scipy import ndimage
import scipy.ndimage.filters as filters
from skimage.morphology import watershed
import numpy as np

from .. import malis
from .. import utils

logger = logging.getLogger('elektronn2log')

rand_index = malis.compute_V_rand_N2  # small is better

def make_affinities(labels, nhood=None, size_thresh=1):
    """
    Construct an affinity graph from a segmentation (IDs)

    Segments with ID 0 are regarded as disconnected
    The spatial shape of the affinity graph is the same as of seg_gt.
    This means that some edges are are undefined and therefore treated as disconnected.
    If the offsets in nhood are positive, the edges with largest spatial index are undefined.

    Connected components is run on the affgraph to relabel the IDs locally.

    Parameters
    ----------

    labels: 4d np.ndarray, int (any precision)
        Volumes of segmentation IDs (bs, z, y, x)
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    size_thresh: int
        Size filters for connected components, smaller objects are mapped to BG


    Returns
    -------

    aff: 5d np.ndarray int16
        Affinity graph of shape (bs, #edges, x, y, z)
        1: connected, 0: disconnected
    seg_gt:
        4d np.ndarray int16
        Affinity graph of shape (bs, x, y, z)
        Relabelling of components
    """

    if nhood is None:
        nhood = np.eye(3, dtype=np.int32)

    aff_sh = [labels.shape[0], nhood.shape[0],] + list(labels.shape[1:])
    out_aff = np.zeros(aff_sh,       dtype=np.int16)
    out_seg = np.zeros(labels.shape, dtype=np.int16)
    for i,l in enumerate(labels):
        out_aff[i] = malis.seg_to_affgraph(l, nhood)
        # we throw away the seg sizes
        out_seg[i], _ = malis.affgraph_to_seg(out_aff[i], nhood, size_thresh)
    return out_aff, out_seg


def make_nhood_targets(target, nhood):
    # nhood (edges, displacements) e.g. (5,3)
    if target.ndim==4:
        raise NotImplementedError
    else:
        assert target.ndim==5
        assert target.shape[1]==1
        sh = target.shape
        new_sh = list(sh)
        new_sh[1] = len(nhood)
        new_target = -1 * np.ones(new_sh, dtype=target.dtype)
        sh = sh[2:]
        for i,off in enumerate(nhood):
            tmp = target[:,0,
                  max(0,-off[0]):sh[0]-max(0,off[0]),
                  max(0,-off[1]):sh[1]-max(0,off[1]),
                  max(0,-off[2]):sh[2]-max(0,off[2])]
            if tmp.size:
                new_target[:, i, max(0,off[0]):sh[0]-max(0,-off[0]),
                                 max(0,off[1]):sh[1]-max(0,-off[1]),
                                 max(0,off[2]):sh[2]-max(0,-off[2])] = tmp

    return new_target



def downsample_xy(d, l, factor):
    """
    Downsample by averaging
    :param d: data
    :param l: label
    :param factor:
    :return:
    """
    f     = int(factor)
    l_sh  = l.shape
    cut   = np.mod(l_sh, f)

    d     = d[:, :, :l_sh[-2]-cut[-2], :l_sh[-1]-cut[-1]]
    sh    = d[:, :, ::f, ::f].shape
    new_d = np.zeros(sh, dtype=np.float32)

    l     = l[:, :, l_sh[-2]-cut[-2], :l_sh[-1]-cut[-1]]
    sh    = l[:, :, :f, ::f].shape
    new_l = np.zeros(sh, dtype=l.dtype)

    for i in range(f):
        for j in range(f):
            new_d += d[:, :, i::f, j::f]
            new_l += l[:,    i::f, j::f]

    d = new_d / f**2
    l = new_l / f**2

    return d, l

@utils.timeit
@numba.jit(nopython=True)
def _ids2barriers(ids, barriers, dilute, connectivity):
    """
    Draw a 2 or 4 pix barrier where label IDs are different

    :param ids:  (x,y,z)
    :param barriers:
    :param dilute: e.g. [False, True, True]
    :param connectivity: e.g. [True, True, True]
    :return:
    """
    nx = ids.shape[0]
    ny = ids.shape[1]
    nz = ids.shape[2]

    for x in np.arange(nx-1):
        for y in np.arange(ny-1):
            for z in np.arange(nz-1):
                if connectivity[0]:
                    if ids[x,y,z]!=ids[x+1,y,z]:
                        barriers[x,y,z]   = 1
                        barriers[x+1,y,z] = 1
                        if dilute[0]:
                            if x>0:    barriers[x-1,y,z] = 1
                            if x<nx-2: barriers[x+2,y,z] = 1

                if connectivity[1]:
                    if ids[x,y,z]!=ids[x,y+1,z]:
                        barriers[x,y,z]   = 1
                        barriers[x,y+1,z] = 1
                        if dilute[1]:
                            if y>0:    barriers[x,y-1,z] = 1
                            if y<ny-2: barriers[x,y+2,z] = 1

                if connectivity[2]:
                    if ids[x,y,z]!=ids[x,y,z+1]:
                        barriers[x,y,z]   = 1
                        barriers[x,y,z+1] = 1
                        if dilute[2]:
                            if z>0:    barriers[x,y,z-1] = 1
                            if z<nz-2: barriers[x,y,z+2] = 1




def ids2barriers(ids, dilute=[True,True, True],
                 connectivity=[True, True, True],
                 ecs_as_barr=True,
                 smoothen=False):
    dilute = np.array(dilute)
    connectivity = np.array(connectivity)
    barriers = np.zeros_like(ids, dtype=np.int16)

    _ids2barriers(ids, barriers, dilute, connectivity)
    _ids2barriers(ids[::-1,::-1,::-1],
                  barriers[::-1,::-1,::-1],
                  dilute, connectivity) # apply backwards as lazy hack to fix boundary
                  
    if smoothen:
        kernel = np.array([[[0.1, 0.2, 0.1],
                            [0.2, 0.3, 0.2],
                            [0.1, 0.2, 0.1]],

                           [[0.3, 0.5, 0.3],
                            [0.5, 1.0, 0.5],
                            [0.3, 0.5, 0.3]],

                           [[0.1, 0.2, 0.1],
                            [0.2, 0.3, 0.2],
                            [0.1, 0.2, 0.1]]])
                            
        barriers_s = filters.convolve(barriers.astype(np.float32),
                                      kernel.astype(np.float32))
        barriers = (barriers_s>4).astype(np.int16) # (old - new).mean() ~ 0

    if ecs_as_barr=='new_class':
        ecs  = np.logical_and( (ids==0), (barriers!=1))
        barriers[ecs] = 2
    
    elif ecs_as_barr:
        ecs  = (ids==0).astype(np.int16)
        barriers = np.maximum(ecs, barriers)

    return barriers

def blob(sizes):
    """
    Return Gaussian blob filter
    """
    grids = [np.linspace(-2.2,2.2,size) for size in sizes]
    grids = np.meshgrid(*grids, indexing='ij')
    ret = np.exp(-0.5*(reduce(np.add, list(map(np.square, grids)))))
    ret = ret / np.square(ret).sum()
    return ret

def _smearbarriers(barriers, kernel):
    # Note: this is good but makes holes to small,
    #  besides we must raise/lower all confidences in GT
    barriers = barriers.astype(np.float32)
    if kernel is None:
        kernel = np.array([
            [[ 0.,  0.,  0.,  0.,  0.],
             [ 0.,  0.,  0.1,  0.,  0.],
             [ 0.,  0.1,  0.2,  0.1,  0.],
             [ 0.,  0.,  0.1,  0.,  0.],
             [ 0.,  0.,  0.,  0.,  0.]],

            [[ 0.,  0.,  0.1,  0.,  0.],
             [ 0.,  0.2,  0.4,  0.2,  0.],
             [ 0.1,  0.4,  1.,  0.4,  0.1],
             [ 0.,  0.2,  0.4,  0.2,  0.],
             [ 0.,  0.,  0.1,  0.,  0.]],

            [[ 0.,  0.,  0.,  0.,  0.],
             [ 0.,  0.,  0.1,  0.,  0.],
             [ 0.,  0.1,  0.2,  0.1,  0.],
             [ 0.,  0.,  0.1,  0.,  0.],
             [ 0.,  0.,  0.,  0.,  0.]],
        ]).T
    else:
        sizes = kernel
        kernel = blob(sizes)
        index = np.subtract(sizes, 1)
        index = np.divide(index, 2)
        kernel[tuple(index)] = 1.0 # set center to 1


    barriers = filters.convolve(barriers, kernel)
    barriers = np.minimum(barriers, 1.0)
    return barriers

def smearbarriers(barriers, kernel=None):
    """
    barriers: 3d volume (z,x,y)
    """
    pos = _smearbarriers(barriers, kernel)
    neg = 1.0 - _smearbarriers(1.0 - barriers, kernel)
    barriers = 0.5 * (pos + neg)
    #barriers = np.minimum(barriers, 1.0)
    return barriers


@numba.jit(nopython=True)
def _grow_seg(seg, grow, mask):
    nx = seg.shape[0]
    ny = seg.shape[1]
    nz = seg.shape[2]
    for x in range(1,nx-1):
        for y in range(1,ny-1):
            for z in range(1,nz-1):

                if mask[0] and (seg[x,y,z]!=0) and (seg[x-1,y,z]==0):
                    grow[x-1,y,z]   = seg[x,y,z]
                if mask[0] and (seg[x,y,z]!=0) and (seg[x+1,y,z]==0):
                    grow[x+1,y,z]   = seg[x,y,z]

                if mask[1] and (seg[x,y,z]!=0) and (seg[x,y-1,z]==0):
                    grow[x,y-1,z]   = seg[x,y,z]
                if mask[1] and (seg[x,y,z]!=0) and (seg[x,y+1,z]==0):
                    grow[x,y+1,z]   = seg[x,y,z]

                if mask[2] and (seg[x,y,z]!=0) and (seg[x,y,z-1]==0):
                    grow[x,y,z-1]   = seg[x,y,z]
                if mask[2] and (seg[x,y,z]!=0) and (seg[x,y,z+1]==0):
                    grow[x,y,z+1]   = seg[x,y,z]

def grow_seg(seg, pixel=[1,3,3]):
    """
    Grow segmentation labels into ECS/background by n pixel
    """
    if isinstance(pixel, (list, tuple, np.ndarray)):
        n = np.max(pixel)
    else:
        n = pixel
        pixel = [n,] * 3

    if n==0:
        return seg

    grow = seg.copy()
    for i in range(n):
        mask = np.greater(pixel, 0)
        _grow_seg(seg, grow, mask)
        seg = grow.copy()
        pixel = np.subtract(pixel, 1)

    return seg



def center_cubes(cube1, cube2, crop=True):
    """
    shapes (ch,x,y,z) or (x,y,z)
    """
    is_3d = [False, False]
    if cube1.ndim==3:
        cube1 = cube1[None]
        is_3d[0] = True
    if cube2.ndim==3:
        cube2 = cube2[None]
        is_3d[1] = True

    diffs = np.subtract(cube1.shape, cube2.shape)[1:]
    assert np.all(diffs%2==0)
    diffs //= 2

    slices1 = [slice(None)]
    pad1    = [(0,0)]
    slices2 = [slice(None)]
    pad2    = [(0,0)]
    for d in diffs:
        if d>0: # 1 is larger than 2
            if crop:
                slices1.append(slice(d, -d))
                pad1.append((0,0))

                slices2.append(slice(None))
                pad2.append((0,0))
            else:
                slices1.append(slice(None))
                pad1.append((0,0))

                slices2.append(slice(None))
                pad2.append((d,d))
        elif d<0:
            if crop:
                slices2.append(slice(-d, d))
                pad2.append((0,0))

                slices1.append(slice(None))
                pad1.append((0,0))
            else:
                slices2.append(slice(None))
                pad2.append((0,0))

                slices1.append(slice(None))
                pad1.append((-d, -d))
        else:
            slices2.append(slice(None))
            pad2.append((0,0))

            slices1.append(slice(None))
            pad1.append((0,0))

    cube1 = cube1[slices1]
    cube2 = cube2[slices2]
    cube1 = np.pad(cube1, pad1, 'constant')
    cube2 = np.pad(cube2, pad2, 'constant')

    if is_3d[0]:
        cube1 = cube1[0]
    if is_3d[1]:
        cube2 = cube2[0]

    return cube1, cube2

### Segmentation ##############################################################
def seg_old(pred=None, thresh=10.0, hi_thresh=140, lo_thresh=18,
            grow_it=4, slack_dt=True, scale = [20,9,9]):
    """
    pred: int prob map (z,x,y) !!!!!
    thresh: threshold for
    """

    # This with high threshold -> more holes
    mem_high = np.invert(((pred > hi_thresh) * 255).astype(np.uint8))
    dt_objects_ws     = -ndimage.distance_transform_edt(mem_high, sampling=scale)

    # This is with low threshold -> less holes, closes small stuff
    mem_low = np.invert(((pred > lo_thresh) * 255).astype(np.uint8))
    dt_objects_labels = -ndimage.distance_transform_edt(mem_low, sampling=scale)
    # Both distance transforms have 0 on the "hard membrane" and large negative values inside segments


    if slack_dt:
        smem = ((pred > hi_thresh) * 255).astype(np.uint8)
        dt_slack = -ndimage.distance_transform_edt(smem, sampling=scale)
        dt_comb = dt_slack + dt_objects_ws
        del dt_slack

    else:
        dt_comb = dt_objects_ws

    # create a slack label, this is label 1
    # regions where membrane/background/ecs is thick
    slack_label = ndimage.morphology.binary_erosion(pred > 150, iterations=3) * 1

    seeds, num = ndimage.measurements.label(dt_objects_labels<-thresh) # CC
    seeds[seeds!=0] += 1 # "make space" for the slack label seed with ID 1
    #print "SHIFT ON!"
    seeds[slack_label==1] = 1
    ws = watershed(dt_comb, seeds) -1
    ws = ws.astype(np.int16)
    ws = grow_seg(ws, pixel=grow_it)

    return ws

def seg_proc(kwargs):
    gt = kwargs.pop('gt')
    seg = seg_old(**kwargs)
    kwargs.pop('pred')
    ri, ri_split, ri_merge = rand_index(gt, seg)

    print("RI=%.4f\tthresh=%i\thi_thresh=%i\tlo_thresh=%i\tgrow=%s"\
          %(ri, kwargs['thresh'], kwargs['hi_thresh'], kwargs['lo_thresh'],
            kwargs['grow_it']))

    return ri

def optimise_segmentation(gt, pred, save_name, n_proc=2):
    threshs = [37,40,43,56] # 4
    hi_threshs = [235,240,245] # 4
    lo_threshs = [1,2,3] # 4
    grow_its   = [(0,0,0),(1,3,3),(2,6,6),(3,6,6)] # 3

    args = []
    for thresh in threshs:
        for grow_it in grow_its:
            for hi_thresh in hi_threshs:
                for lo_thresh in lo_threshs:
                    args.append(dict(gt=gt,
                                     pred=pred,
                                     thresh=thresh,
                                     hi_thresh=hi_thresh,
                                     lo_thresh=lo_thresh,
                                     grow_it=grow_it))

    print("Scanning for best SEGMENTATION parameters")
    mp = multiprocessing.Pool(n_proc)
    rand_indices  = list(mp.map(seg_proc, args))
    rand_indices = np.array(rand_indices)

    for i in range(len(args)):
        args[i].pop('gt')

    min_i = np.argmin(rand_indices)
    seg = seg_old(**args[min_i])
    best_ri = rand_indices[min_i]
    for i in range(len(args)):
        args[i].pop('pred')

    report_str = "%s Evaluation\n"%(save_name)
    report_str += "BEST: RI=%.4f "%rand_indices[min_i] + str(args[min_i]) + '\n'


    for re, config in zip(rand_indices, args):
        report_str += "RI=%.4f "%re + str(config) + '\n'

    with open("%s-Seg_Params.txt" %(save_name), 'w') as f:
        f.write(report_str)

    utils.h5save(randomise_colours(seg), '%s_seg.h5' %(save_name,), 'seg')
    print("Best RI=%.4f" %(best_ri))

    return rand_indices, best_ri, seg

def randomise_colours(a, size_filter=1500):
    out = np.zeros_like(a)
    bc = np.bincount(a.ravel())
    colors = (bc>size_filter).nonzero()[0]
    new_cols = np.random.permutation(len(colors)-1)
    new_cols   = np.hstack([0, new_cols])
    for c in colors:
        i = np.argmax(colors==c)
        out[a==c] = new_cols[i]

    return out


def billig_seg(gt, pred, thresh, ecs_thresh):
    seeds, num = ndimage.measurements.label((pred<thresh))
    ws = watershed(pred, seeds)
    ws[pred>ecs_thresh] = 0
    seg = ws
    ri, ri_split, ri_merge = rand_index(gt, seg)
    return ri, seg


def billig_seg_proc(kwargs):
    gt = kwargs['gt']
    pred = kwargs['pred']
    thresh = kwargs['thresh']
    ecs_thresh = kwargs['ecs_thresh']
    ri, seg = billig_seg(gt, pred, thresh, ecs_thresh)
    print("RI=%.4f\tthresh=%i\tecs_thresh=%i"%(ri, thresh, ecs_thresh))
    return ri

def optimise_billig_segmentation(gt, pred, save_name, n_proc=2):
    threshs = np.linspace(100, 240, 6) # 4
    ecs_threshs = [253,] #np.linspace(230, 255,5) # 4

    args = []
    for thresh in threshs:
        for ecs_thresh in ecs_threshs:
            args.append(dict(gt=gt,
                             pred=pred,
                             thresh=thresh,
                             ecs_thresh=ecs_thresh))

    print("Scanning for best SEGMENTATION parameters")
    mp = multiprocessing.Pool(n_proc)
    rand_indices  = list(mp.map(billig_seg_proc, args))
    rand_indices = np.array(rand_indices)

    min_i = np.argmin(rand_indices)
    a = args[min_i]
    ri, seg = billig_seg(a['gt'], a['pred'], a['thresh'], a['ecs_thresh'])

    for i in range(len(args)):
        args[i].pop('pred', None)
        args[i].pop('gt', None)

    report_str = "%s Evaluation\n"%(save_name)
    report_str += "BEST: RI=%.4f "%rand_indices[min_i] + str(args[min_i]) + '\n'

    for re, config in zip(rand_indices, args):
        report_str += "RI=%.4f "%re + str(config) + '\n'

    with open("%s-Seg_Params.txt" %(save_name), 'w') as f:
        f.write(report_str)

    utils.h5save(randomise_colours(seg), '%s_seg.h5' %(save_name,), 'seg')
    print("Best RI=%.4f" %(rand_indices[min_i]))

    return rand_indices[min_i], seg



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


__all__ = ['Node', 'Input', 'Input_like', 'Concat', 'ApplyFunc',
           'FromTensor', 'split', 'model_manager', 'GenericInput', 'ValueNode',
           'MultMerge', 'InitialState_like']


import sys
import inspect
import logging
import re
import time
import functools
import uuid
import gc
from collections import OrderedDict
from functools import reduce

import numpy as np
import theano
import theano.tensor as T
import tqdm

from theano import gof
from theano import printing
from future.utils import raise_with_traceback
from future.utils import with_metaclass

from .variables import VariableWeight, ConstantParam
from .. import utils
from . import graphutils

if sys.version_info.major >= 3:
    unicode = str

logger = logging.getLogger('elektronn2log')

floatX = theano.config.floatX

###############################################################################

class ModelContainer(object):
    """
    Keeps track of the network models.

    This class describes the global model_manager, which can contain multiple models that
    were defined by a network configuration.
    Each model is represented by a Model object.
    """
    _count = 0

    def __init__(self):
        if self._count > 0:
            raise RuntimeError("There may exist only one ModelContainer ("
                               "which may contain several models)")

        self._count += 1
        self._default_set = False
        self.models = OrderedDict()
        self.last = None
        self.current = None


    def __getitem__(self, item):
        return self.models[item]

    def __repr__(self):
        return repr(list(self.models.keys()))

    def setdefault(self):
        """
        If it is not already set, a new model named "default" is created and
        registered as the default model.
        """
        if self._default_set:
            raise RuntimeError("The default model has already been set.")
        from elektronn2.neuromancer.model import Model
        self.models["default"] = Model(name="default")
        self.current = self["default"]
        self.last = self["default"]
        self._default_set = True

    def newmodel(self, name):
        """
        Create a new Model in the ModelContainer and return the new Model
        object.

        Parameters
        ----------
        name: str
            Name of the new Model. If a Model with this name exists,
            a random UUID is used instead.

        Returns
        -------
        Model
            The new Model object.
        """
        from elektronn2.neuromancer.model import Model
        if name in self.models:
            raise ValueError("Model of the same name %s exists already." %(name))
        elif name is None:
            name = str(uuid.uuid4())
        self.models[name] = Model(name=name)
        self.last = self.current
        self.current = self[name]
        return self[name]

    def getmodel(self, *args):
        """
        Return a Model and register it as the current one.

        Parameters
        ----------
        *args
            If empty, the default Model is returned. If it is a string, the
            model with this name is returned.

        Returns
        -------
        Model
            The Model object.
        """
        if len(args)==1:
            current = self[args[0]]
        elif len(args)==0:
            current =  self["default"]
        else:
            raise ValueError("Either provide name or nothing!")

        self.last = self.current
        self.current = current
        return current

    def togglemodel(self):
        """
        Swap last and current model.

        :return: current Model
        """
        self.last, self.current = self.current, self.last
        return self.current

# Create a quasi-global, unique GraphManagerContainer and bind it the node
# registration decorator. This implies a mutable state, because the
model_manager = ModelContainer()

def choose_name(proposal, names):
    """
    Choose an appropriate unique name for a Node.

    If the proposed name is not already taken, it is directly returned.
    If it is taken and it does not end with a number, a "1" is appended to it.
    If it ends with a number and is taken, the number is incremented by 1
    until the proposed name is unique.

    Parameters
    ----------
    proposal: str
        Proposed name for the Node.
    names: str
        Names that were already given to other nodes in this Model.

    Returns
    -------
    str
        Appropriate and unique name for the Node.

    Examples
    --------
    This function makes it possible to write a network config in the following style:
    >>> out = neuromancer.Conv(inp, 12,  (3,3), (2,2))
    >>> out = neuromancer.Conv(out, 36,  (3,3), (2,2))
    >>> out = neuromancer.Conv(out, 64,  (3,3), (1,1))
    , which will automatically assign the names "conv", "conv1" and "conv2" to
    the three subsequent Conv nodes, since the proposed default name for Conv
    nodes is "conv".
    """
    if proposal in names:
        numbers = re.findall('(\d+)$', proposal)
        if not len(numbers):
           proposal = proposal+str(1)

        while proposal in names:
            proposal = re.sub('(\d+)$', lambda x: str(int(x.group(0)) + 1), proposal)

    return proposal


class MetaNode(type):
    """
    Metaclass for the ``Node`` class that manages automatic registration in the
    current ``model_manager``.

    This metaclass makes sure that the global ``model_manager``, which keeps track
    of the network models, is updated every time a new ``Node`` object is created.
    """
    def __new__(mcs, name, bases, clsdict):
        if name=='Node': # Don't do anything to base class
            pass
        else:
            for attr, attribute in clsdict.items():
                if attr == "__init__":
                    attribute = mcs.init_register(attribute)
                if attr == "__new__":
                    pass

                # Inherit class docstring
                if not attribute.__doc__:
                    for mro_cls in (mro_cls for base in bases for mro_cls in
                                    base.mro()
                                    if hasattr(mro_cls, attr)):
                        doc = getattr(getattr(mro_cls, attr), '__doc__')
                        if doc:
                            attribute.__doc__ = doc+"        NOTE: docstring was inherited\n"
                            break

                clsdict[attr] = attribute

        return super(MetaNode, mcs).__new__(mcs, name, bases, clsdict)

    @staticmethod
    def init_register(old_init):
        """
        Wrap the Node class constructor with with model management routines.

        This makes sure that every Node object that is created is registered
        within the current network model.

        Parameters
        ----------
        old_init: function
            Original constructor of the Node

        Returns
        -------
        function
            New constructor that wraps the original one
        """
        @functools.wraps(old_init)
        def new_init(*args, **kwargs):
            if model_manager.current is None:
                n_pre = 0
            else:
                n_pre = len(model_manager.current.nodes)

            node = args[0]
            args = args[1:]
            if model_manager.current is None:
                model_manager.setdefault()

            default_name = ''
            arg_names, varargs, keywords, defaults = inspect.getargspec(old_init)
            try: # extract the default name from kwargs of class initialiser
                i = arg_names.index('name')
                i = i - len(arg_names) + len(defaults)
                default_name = defaults[i]
                assert isinstance(default_name, str)
            except: # If fails just use an empty default name
                pass

            # Check to register node not twice. This might happen if a node
            # calls an inherited __init__ which has already a register
            # decorator, but the actual __init__ has already done the
            # registration. It is important that the registration is done
            # prior to the call to __init__!
            if not node in model_manager.current.nodes.values():
                # Replace given name by unique identifier (if not unique)
                name = kwargs.get('name', default_name)
                name = choose_name(name, model_manager.current.node_descriptors.keys())
                kwargs['name'] = name
                model_manager.current.register_node(node, name, args, kwargs)

            old_init(node, *args, **kwargs)
            node._finalize_init()

            n_post = len(model_manager.current.nodes)
            diff = n_post - n_pre
            if diff > 1:
                raise RuntimeError("Initialisation of %s failed: instead of 1"
                                   " %i were registered to the model_manager. It is"
                                   "not permitted to create other nodes within"
                                   "the class definiton of a node. Use a factory"
                                   "function to create several nodes at once"
                                   %(args[0].__class__, diff))
            elif diff == 0:
                logger.debug("Initialisation of %s registered 0 instead of 1 nodes. "
                            "This might be the case, if this function is "
                            "called from a derived node that has already registered."
                                   % (args[0].__class__,))

        return new_init


###############################################################################

class Node(with_metaclass(MetaNode, object)):
    """
    Basic node class. All neural network nodes should inherit from ``Node``.

    Parameters
    ----------
    parent: Node or list[Node]
        The input node(s).
    name: str
        Given name of the Node, may be an empty string.
    print_repr: bool
        Whether to print the node representation upon initialisation.


    Models are built from the interplay of *Nodes* to form a (directed,
    acyclic) computational graph.

    The **ELEKTRONN2** framework can be seen as an intelligent abstraction level
    that hides the raw theano-graph and manages the involved symbolic variables.
    The overall goal is the intuitive, flexible and **easy** creation of
    complicated graphs.

    A ``Node`` has one or several inputs, called ``parent``, (unless it is
    a *source*, i.e. a node where external data is feed into the graph).
    The inputs are node objects themselves.

    Layers automatically keep track of their previous inputs,
    parameters, computational cost etc. This allows to compile the
    theano-functions without manually specifying the inputs, outputs and
    parameters.
    In the most simple case, any node, which might be part of a more
    complicated graph, can be called as a function (passing suitable
    numpy arrays):

     >>> import elektronn2.neuromancer.utils
     >>> inp = neuromancer.Input((batch_size, in_dim))
     >>> test_data = elektronn2.neuromancer.utils.as_floatX(np.random.rand(batch_size, in_dim))
     >>> out = inp(test_data)
     >>> np.allclose(out, test_data)
     True

    At the first time the theano function is compiled and cached
    for re-use in future calls.

    Several properties (with respect to the sub-graph the node depends on, or
    only from the of the node itself)
    These can also be looked up externally (e.g. required sources,
    parameter count, computational count).

    The theano variable that represents the output of a node is kept
    in the attribute ``output``. Subsequent Nodes must use this attribute of
    their inputs to perform their calculation and write the result to their own
    output (this happens in the method ``_calc_output``, which is hidden
    because it must be called only internally at initialisation).

    A divergence in the computational graph is created by passing the *parent*
    to several *children* as input:

     >>> inp = neuromancer.Input((1,10), name='Input_1')
     >>> node1 = neuromancer.ApplyFunc(inp, func1)
     >>> node2 = neuromancer.ApplyFunc(inp, func2)


    A convergence in the graph is created by passing several inputs to a node
    that performs a reduction:

     >>> out = neuromancer.Concat([node1, node2])

    Although the node "out" has two immediate inputs, it is detected that the
    required sources is only a single object:

      >>> print(out.input_nodes)
      Input_1

    Computations that result in more than a single output for a Node must be
    broken apart using divergence and individual nodes for the several outputs.
    Alternatively the function ``split`` can be used to create two
    dummy nodes of the output of a previous Node by splitting along the
    specified axis.
    Note that possible redundant computations in Nodes are most likely
    eliminated by the theano graph optimiser.


    **Instructions for subclassing:**

    Overriding ``__init__``:
        At the very first the base class' initialiser must be called, which just
        assigns the names and emtpy default values for attributes.
        Then node specific initialisations are made e.g. initialisation of shared
        parameters / weights.
        Finally the ``_finialise_init`` method of the base class is automatically
        called:
        This evokes the execution of the methods: ``_make_output``,
        ``_calc_shape`` and ``self._calc_comp_cost``. Each of those updates
        the corresponding attributes.
        NOTE: if a Node (except for the base ``Node``) is subclassed and the
        derived calls ``__init__`` of the base Node, this will also call
        ``_finialise_init`` exactly right the call to the superclass'
        ``__init__``.

    For the graph serialisation and restoration to work, the following
    conditions must additionally be met:

    * The name of of a node's trainable parameter in the parameter dict must
      be the same as the (optional) keyword used to initialise this parameter
      in ``__init__``;
      moreover, parameters must not be initialised/shared from positional
      arguments.
    * When serialising only the current state of parameters is kept, parameter
      value arrays given for initialisation are never kept.

    Depending on the purpose of the node, the latter methods and others
    (e.g. ``__repr__``) must be overridden. The default behaviour of the base
    class is: output = input, outputs shape = input shape,
    computational cost = tensor size (!) ...
    """

    def __init__(self, parent, name="", print_repr=False):
        self.parent             = parent
        self.children           = OrderedDict()
        self.name               = name
        self._features_names    = None
        # weights, Dropout rate, mixing rates, batch normalisation etc.
        self.params             = OrderedDict()
        self.computational_cost = 0

        self.is_source   = False
        self.output       = None # theano tensor variable
        self.shape        = None # The *output* shape
        self._output_func = None # theano function (compiled on demand)
        self._debug_outputs = []
        self._output_func_debug = None
        self._local_exec_time = None
        self._total_exec_time = None
        self._prof = None

        self._finalized = False
        self._print_repr = print_repr


    def __repr__(self):
        if self.name=='':
            s  = "<%s-Node>\n"%(self.__class__.__name__,)
        else:
            s  = "<%s-Node> '%s' \n"%(self.__class__.__name__, self.name)
        s += '  '
        if self.param_count>0:
            s += "#Params={0:,d} ".format(self.param_count)
        if self.computational_cost>0:
            s += "Comp.Cost=%s, " %(utils.pretty_string_ops(self.computational_cost))
        s += "Out:%s" %(str(self.shape))
        if len(self.input_nodes)>1:
            sources = [node.name for node in self.input_nodes]
            s+= "\n  Order of sources=%s, "%(str(sources))
        return s


    def __call__(self, *args):
        """
        Compute output of node. If called for the first time the theano-function
        must be compiled which might take a while.

        Parameters
        ----------

        *args: numpy arrays
            Graph input

        Returns
        -------
        np.ndarray
            Node output.
        """
        try:
            return self._output_func(*args)
        except TypeError as e:
            tb = sys.exc_info()[2]
            if len(args)!=len(self.input_nodes):
                add_info = "\n %i inputs required, %i were given." %(len(self.input_nodes), len(args))
            else:
                try:
                    add_info = '\nShapes (required - given):\n'
                    for ni, ar in zip(self.input_nodes, args):
                        add_info += " %s %s\t%s %s\n" %(tuple(ni.shape.shape), ni.dtype, ar.shape, ar.dtype)
                except:
                    add_info = ""

            raise_with_traceback(type(e)(str(e) + add_info), tb)

    def get_debug_outputs(self, *args):
        if self._output_func_debug:
            return self._output_func_debug(*args)
        else:
            logger.warning("Layer '%s' has no debug outputs defined." %(self.name,))

    def _finalize_init(self):
        """
        Calculates output, computational cost and shape, prints node repr (if
        not turned off)
        """
        if self._finalized:
            logger.debug("Attempt to call _finalize_init twice on %s. This "
            "might be the case, if this function is called from a derived node "
            "and the base has already called it. This call is IGNORED." %self.name)
            return

        # Don't change the order of this!
        self._make_output()
        self._calc_shape()
        self._calc_comp_cost()
        if isinstance(self.parent, (list, tuple)):
            for p in self.parent:
                p._register_child(self)
                # logger.debug('{} {} {}'.format(id(p), p.name, p.children.keys())
        else:
            if self.parent is not None:
                self.parent._register_child(self)

        self._output_func = graphutils.make_func(self.input_tensors,
                                                 self.output, name=self.name)

        if self._debug_outputs:
            self._output_func_debug = graphutils.make_func(self.input_tensors,
                                                     self._debug_outputs, name=self.name+"_dbg")

        if self._print_repr:
            logger.info("-" * 87)
            logger.info(self)
        else:
            logger.debug("-" * 87)
            logger.debug(self)

        self._finalized = True

    def _register_child(self, child):
        """
        Add child to dict of children
        """
        self.children[child.name] = child
        #print("Adding Child %s to node %s"%(child.name, self.name))


    def _make_output(self):
        """
        Computation of output variable (symbolic)
        """
        self.output = self.parent.output


    def _calc_shape(self):
        """
        Computation of output shape and tags
        """
        if isinstance(self.parent, (list, tuple)):
            self.shape = self.parent[0].shape.copy()
        else:
            self.shape = self.parent.shape.copy()


    def _calc_comp_cost(self):
        """
        Computation of computational cost
        """
        if isinstance(self.parent, (list, tuple)):
            self.computational_cost = self.parent[0].shape.stripnone_prod
        else:
            self.computational_cost = self.parent.shape.stripnone_prod


    def get_param_values(self, skip_const=False):
        """
        Returns a dict that maps the values of the params.
        (such that they can be saved to disk)

        Parameters
        ----------
        skip_const: bool
            whether to exclude constant parameters.

        Returns
        -------
        dict
            Dict that maps the values of the params.
        """
        p_dict = OrderedDict()
        for k,v in self.params.items():
            if v.constant and skip_const:
                logger.debug("Skipping %s because it is constant" % (k,))
            else:
                p_dict[k] = v.get_value()

        return p_dict

    def set_param_values(self, value_dict, skip_const=False):
        """
        Sets new values for non constant parameters.

        Parameters
        ----------
        value_dict: dict
            A dict that maps values by parameter name.
        skip_const: bool
            if dict also maps values for constants, these can be skipped,
            otherwise an exception is raised.
        """
        for k,v in value_dict.items():
            if k not in self.params:
                if k in ['gamma', 'std', 'mean']:
                    logger.info("Skipping set_value of BN parameter %s" %(k,))
                    continue
                raise KeyError("Layer has no parameter %s" %(k,))

            if self.params[k].constant:
                if skip_const:
                    logger.debug("Skipping %s because it is constant" % (k,))
                    continue
                else:
                    raise ValueError("Cannot set value of constant "
                                     "parameter %s in node %s" %(k, self.name))

            self.params[k].set_value(v)

    @property
    def all_parents(self):
        """
        List all nodes that are involved in the computation of the output
        of this node (incl. ``self``). The list contains no duplicates.
        The return is a dict, the keys of which are the layers, the values are
        just all ``True``
        """
        parents = OrderedDict()

        if self.parent is None: # For an input/source node
            pass
        elif isinstance(self.parent, (tuple, list)):
            for node in self.parent:
                parents.update(node.all_parents)
        else:
            parents.update(self.parent.all_parents)

        parents[self.name] = self # insert this node itself
        return parents


    @property
    def input_nodes(self):
        """
        Contains the all parent nodes that are sources, i.e. inputs that are
        required to compute the result of this node.
        """
        inputs = []
        for node in self.all_parents.values(): # this does not contain duplicates
            if node.is_source:
                inputs.append(node)
        return inputs

    @property
    def input_tensors(self):
        """
        The same as ``input_nodes`` but contains the theano tensor variables
        instead of the node objects. May be used as input to compile theano
        functions.
        """
        return [s.output for s in self.input_nodes]

    @property
    def all_params(self):
        """
        Dict of the all parameters of all parent nodes. They are theano variable
        """
        p_dict = OrderedDict()
        for node in self.all_parents.values():
            for k,v in node.params.items():
                p_dict[str(node.name)+'_'+k] = v
        return p_dict


    @property
    def all_trainable_params(self):
        """
        Dict of the trainable parameters (weights) of all parent nodes. They
        are theano shared variables.
        """
        p_dic = OrderedDict()
        for node in self.all_parents.values():
            for k,v in node.params.items():
                if v.apply_train:
                    p_dic[str(node.name)+'_'+k] = v
        return p_dic


    @property
    def all_nontrainable_params(self):
        """
        Dict of the trainable parameters (weights) of all parent nodes. They
        are theano shared variables.
        """
        p_dic = OrderedDict()
        for node in self.all_parents.values():
            for k,v in node.params.items():
                if not v.apply_train:
                    p_dic[str(node.name)+'_'+k] = v
        return p_dic

    @property
    def all_extra_updates(self):
        """
        List of the parameters updates of all parent nodes. They
        are tuples.
        """
        updates = []
        for node in self.all_parents.values():
            for param in node.params.values():
                if param.updates:
                    updates.append(param.updates)
        return updates

    @property
    def param_count(self):
        """
        Count of trainable parameters in this node
        """
        if self.params:
            param_count = reduce(lambda x,y: x+y.get_value().size,
                                 [x for x in self.params.values() if x.apply_train],
                                 0)
        else:
            param_count = 0

        return param_count

    @property
    def all_params_count(self):
        """
        Count of all trainable parameters in the entire sub-graph used to
        compute the output of this node
        """
        if self.all_trainable_params:
            param_count = reduce(lambda x,y: x+y.get_value().size,
                                 self.all_trainable_params.values(), 0)
        else:
            param_count = 0

        return param_count

    @property
    def all_computational_cost(self):
        return reduce(lambda x,y: x+y.computational_cost,
                      self.all_parents.values(), 0)

    @property
    def all_children(self):
        children = OrderedDict()
        for child in self.children.values():
            children.update(child.all_children)

        return children

    @property
    def last_exec_time(self):
        """
        Last function execution time in seconds.
        """
        return self._output_func.last_exec_time

    @property
    def local_exec_time(self):
        if self._local_exec_time is None:
            self.measure_exectime(print_info=False, local=True)

        return self._local_exec_time

    @property
    def total_exec_time(self):
        if self._total_exec_time is None:
            self.measure_exectime(print_info=False, local=True)

        return self._total_exec_time


    @property
    def feature_names(self):
        if self._features_names:
            return self._features_names
        else:
            return None

    @feature_names.setter
    def feature_names(self, value):
        if 'f' not in self.shape.tags:
            raise ValueError("Shape has no feature tag")
        if len(value)!= self.shape['f']:
            raise ValueError("Shape has %i features, but %i names were"
                             "given" %(self.shape['f'], len(value)))

        self._features_names = tuple(value)


    def _predict_densetile(self, raw_img, out_arr):
        """
        Parameters
        ----------
        raw_img: np.ndarray
          raw image (ch, x, y) or (ch, z, y, x)to be predicted
          The shape must be cnn.patch_size + cnn.output_strides - 1 (elwise)
        out_arr: np.ndarray
          The shape is cnn.patch_size + cnn.mfp_strides - floor(cnn.offset) - 1 (elwise)

        Returns
        -------
        np.ndarray
          Prediction (n_lab, z, y, x).
          The shape is cnn.patch_size + cnn.mfp_strides - floor(cnn.offset) - 1 (elwise)

        """
        strides = self.shape.strides
        patch_size = self.input_nodes[0].shape.spatial_shape

        if np.all(np.equal(strides,1)):
            if self.shape.ndim==2:
                out_arr[:,0] = self(raw_img[None])[0] # (ch,x,y)
            else:
                out_arr[:] = self(raw_img[None])[0] # (ch,z,y,x)

        else:
            for x_off in range(strides[-2]):
                for y_off in range(strides[-1]):
                    if self.shape.ndim==2:
                        cut_img = raw_img[None, :,
                                  x_off: x_off+patch_size[0],
                                  y_off: y_off+patch_size[1]]

                        # insert prob(ch, x, y) into out_arr(ch,z,y,x)
                        out_arr[:,
                        0,
                        x_off::strides[0],
                        y_off::strides[1]] = self(cut_img)[0]

                    elif self.shape.ndim==3:
                        for z_off in range(strides[0]):
                            cut_img = raw_img[None,
                                      :,
                                      z_off: z_off+patch_size[0],
                                      x_off: x_off+patch_size[1],
                                      y_off: y_off+patch_size[2]]

                            out_arr[:,
                            z_off::strides[0],
                            x_off::strides[1],
                            y_off::strides[2]] = self(cut_img)[0]

        return out_arr

    def predict_dense(self, raw_img, as_uint8=False, pad_raw=False):
        """
        Core function that performs the inference

        Parameters
        ----------
        raw_img : np.ndarray
            raw data in the format (ch, (z,) y, x)
        as_uint8: Bool
            Return class proabilites as uint8 image (scaled between 0 and 255!)
        pad_raw: Bool
            Whether to apply padding (by mirroring) to the raw input image
            in order to get predictions on the full imgae domain.

        Returns
        -------
        np.ndarray
            Predictions.
        """

        # determine normalisation depending on int or float type
        if self.shape.ndim<2:
            logger.error("'predict_dense' works only for nodes with 2 or 3 "
                         "spatial axes, this node has shape %s."%(self.shape))
            return None

        if self.shape.ndim==2 and raw_img.ndim==3 and \
                        raw_img.shape[0]!=self.input_nodes[0].shape['f']:
            logger.error("If 3d input is given to 2d CNNs, the first axis must"
                         "contain the features/channel and have size %i. Note"
                         "that also 4d input can be given to 2d CNNs in the "
                         "axis order (ch, z, y, x), where for each z-slice a"
                         "2d prediction image is created"
                         %self.input_nodes[0].shape['f'])
            return None

        offset = self.shape.offsets
        if np.any(np.less(offset, 0)):
            raise ValueError("Cannot predict dense because the CNN contains "
                             "UpConvs which cause unknown FOVs. If you use "
                             "UpConvs you should not need predict dense anyway!")

        if raw_img.dtype in [np.int, np.int8, np.int16, np.int32, np.uint32, np.uint,
                             np.uint8, np.uint16, np.uint32, np.uint32]:
            m = 255
        else:
            m  = 1

        raw_img = graphutils.as_floatX(raw_img) / m

        time_start = time.time()
        strip_z = False
        if raw_img.ndim==3:
            strip_z = True
            raw_img = raw_img[:,None] # add singleton z-channel

        n_lab      = self.shape['f']
        cnn_out_sh = self.shape.spatial_shape
        ps         = self.input_nodes[0].shape.spatial_shape
        strides    = self.shape.strides


        if self.shape.ndim==2:
            cnn_out_sh = np.concatenate([[1,], cnn_out_sh])
            ps         = np.concatenate([[1,], ps])
            strides    = np.concatenate([[1,], strides])
            offset     = np.concatenate([[0,], offset])

        if pad_raw:
            raw_img = np.pad(raw_img,
                             [(0,0),
                              (offset[0],offset[0]),
                              (offset[1],offset[1]),
                              (offset[2],offset[2])],
                             mode='symmetric')

        raw_sh     = raw_img.shape[1:] # only spatial, not channels
        tile_sh = np.add(ps, strides) - 1
        prob_sh = np.multiply(cnn_out_sh, strides)
        prob_arr = np.zeros(np.concatenate([[n_lab,], prob_sh]), dtype=floatX)

        pred_sh = np.array([raw_sh[0]-2*offset[0], raw_sh[1]-2*offset[1], raw_sh[2]-2*offset[2] ])
        if as_uint8:
            predictions = np.zeros(np.concatenate(([n_lab,], pred_sh)), dtype=np.uint8)
        else:
            predictions = np.zeros(np.concatenate(([n_lab,], pred_sh)), dtype=floatX)


        # Calculate number of tiles (in 3d: blocks) that need to be performed
        z_tiles = int(np.ceil(float(pred_sh[0])/prob_sh[0]))
        x_tiles = int(np.ceil(float(pred_sh[1])/prob_sh[1]))
        y_tiles = int(np.ceil(float(pred_sh[2])/prob_sh[2]))
        total_nb_tiles = np.product([x_tiles, y_tiles, z_tiles])
        if self._output_func.func is None:
            self._output_func.compile()

        logger.info("Predicting img %s in %i Blocks: (%i, %i, %i)" \
                    %(raw_img.shape, total_nb_tiles, z_tiles, x_tiles, y_tiles))
        self() # This compiles the function
        pbar = tqdm.tqdm(total=total_nb_tiles, ncols=80, leave=False)
        for z_t in range(z_tiles):
            for x_t in range(x_tiles):
                for y_t in range(y_tiles):
                    # For every z_tile a slice of thickness cnn_out_sh[2] is
                    # collected and then collectively written to the output_data
                    raw_tile = raw_img[:,
                               z_t*prob_sh[0]:z_t*prob_sh[0]+tile_sh[0],
                               x_t*prob_sh[1]:x_t*prob_sh[1]+tile_sh[1],
                               y_t*prob_sh[2]:y_t*prob_sh[2]+tile_sh[2]]

                    this_is_end_tile = False if np.all(np.equal(raw_tile.shape[1:],  tile_sh)) else True

                    if this_is_end_tile: # requires 0-padding
                        right_pad = np.subtract(tile_sh, raw_tile.shape[1:]) # (ch,z,x,y)
                        right_pad = np.concatenate(([0,], right_pad)) # for channel dimension
                        left_pad  = np.zeros(raw_tile.ndim, dtype=np.int)
                        pad_with  = list(zip(left_pad,right_pad))
                        raw_tile  = np.pad(raw_tile, pad_with, mode='constant')

                    if self.shape.ndim==2:
                        # slice from raw_tile(ch,z,x,y) --> (ch,x,y)
                        prob_arr = self._predict_densetile(raw_tile[:,0], prob_arr) # returns (ch,z=1,x,y)
                        prob = prob_arr
                    else:
                        prob_arr = self._predict_densetile(raw_tile, prob_arr)
                        prob = prob_arr

                    if this_is_end_tile: # cut away padded range
                        prob = prob[:, :prob_sh[0]-right_pad[1], :prob_sh[1]-right_pad[2], :prob_sh[2]-right_pad[3]]

                    if as_uint8:
                        prob *= 255

                    predictions[:,
                    z_t*prob_sh[0]:(z_t+1)*prob_sh[0],
                    x_t*prob_sh[1]:(x_t+1)*prob_sh[1],
                    y_t*prob_sh[2]:(y_t+1)*prob_sh[2]] = prob

                    pbar.update()

        pbar.close()
        dtime = time.time() - time_start
        speed = np.product(predictions.shape[1:]) * 1.0/1000000/dtime
        dtime = utils.pretty_string_time(dtime)
        logger.info(" Inference speed: %.3f MB or MPix /s, time %s" \
                    %(speed, dtime))

        if strip_z:
            predictions = predictions[:,0,:,:]

        return predictions

    def test_run(self, on_shape_mismatch='warn', debug_outputs=False):
        """
        Test execution of this Node with random (but correctly shaped) data.

        Parameters
        ----------
        on_shape_mismatch: str
            If this is "warn", a warning is emitted if there is a mismatch
            between expected and calculated output shapes.

        Returns
        -------
            Debug output of the Theano function.
        """
        inp = self.input_nodes
        test_vals = []
        for i in inp:
            sh = [1 if s is None else s for s in i.shape.shape]
            test_vals.append(np.random.rand(*sh).astype(i.dtype))

        if debug_outputs:
            return self.get_debug_outputs(*test_vals)

        self() # compile before timing
        y = self(*test_vals)
        shape_match = tuple(y.shape) == tuple(self.shape.shape)
        if not shape_match:
            if np.prod(y.shape) == self.shape.stripnone_prod:
                shape_match = True

        if shape_match:
            sh_str = "Shapes agree"
        else:
            sh_str = 'Shapes do not agree (Outshape should be %s)' % self.shape.shape

        dtime = self._output_func.last_exec_time
        speed = self.shape.spatial_size * 1.0/1000000/dtime
        #dtime = utils.pretty_string_time(dtime)

        print("Node '%s' compiled and computation tested\n"
              "Outshape: %s\n"
              "Runtime: %g s\n"
              "Speed: %.3f MB or MPix /s\n"
              "%s"%(self.name, y.shape, dtime, speed, sh_str))

        if not shape_match:
            if on_shape_mismatch=='warn':
                logger.warning("Shape of computed output and shape calculated "
                               "by layer definition not match")
        return y

    def plot_theano_graph(self, outfile=None, compiled=True, **kwargs):
        """
        Plot the execution graph of this Node's Theano function to a file.

        If "outfile" is not specified, the plot is saved in "/tmp/<NAME>.png"

        Parameters
        ----------
        outfile: str or None
            File name for saving the plot.
        compiled: bool
            If True, the function is compiled before plotting.
        kwargs
            kwargs (plotting options) that get directly passed to
            theano.printing.pydotprint().
        """
        if outfile is None:
            outfile ='/tmp/%s.png'%self.name
        kwargs['outfile'] = outfile
        if 'var_with_name_simple' not in kwargs:
            kwargs['var_with_name_simple'] = True
        if compiled:
            self()
            printing.pydotprint(self._output_func.func, **kwargs)
        else:
            printing.pydotprint(self.output, **kwargs)

    def measure_exectime(self, n_samples=5, n_warmup=4, print_info=True, local=True, nonegative=True):
        """
        Measure how much time the node needs for its calculation (in milliseconds).

        Parameters
        ----------
        n_samples: int
            Number of independent measurements of which the median is taken.
        n_warmup: int
            Number of warm-up runs before each measurement (not taken into account for median calculation).
        print_info: bool
            If True, print detailed info about measurements while running.
        local: bool
            Only compute exec time for this node by subtracting its parents' times.
        nonegative: bool
            Do not return exec times smaller than zero.

        Returns
        -------
        np.float
            median of execution time measurements.
        """
        # os.environ['CUDA_LAUNCH_BLOCKING'] = '1'
        inp = self.input_nodes
        if not inp:
            logger.info('Node {} has no inputs -> skipping node, giving it zero execution time.'.format(self.name))
            return 0.0
        self._prof = theano.ProfileStats(atexit_print=False)
        self._output_func.func = None
        self._output_func.compile(profile=self._prof)
        test_vals = [np.random.rand(*i.shape.shape).astype(i.dtype) for i in inp]
        for _ in range(n_warmup):  # TODO: Warmup doesn't work this way with the new theano profiling. Warmups are currently fully counted.
            self(*test_vals)

        def noprofile_op(op, unwanted_ops=None):
            if unwanted_ops is None:
                unwanted_ops = ['HostFromGpu', 'GpuFromHost']
            return str(op) in unwanted_ops

        def measure_once():
            self(*test_vals)
            # exec_time = self.last_exec_time  # TODO: Is n_f already taken into account here?
            # exec_time = self._output_func.func.profile.fct_call_time / self._output_func.func.profile.fct_callcount
            exec_time = sum(t for op, t in self._prof.op_time().items()
                            if not noprofile_op(op)
                            ) / self._prof.fct_callcount
            return exec_time * 1000  # in ms

        subgraph_exec_times = np.array([measure_once() for _ in range(n_samples)])
        sort_ix = np.argsort(subgraph_exec_times)
        # TODO: If the profile.op_times() / profile.fct_callcount() approach stays, mean or median don't make sense here
        median_exec_time = subgraph_exec_times[sort_ix[-3:]].mean() #  np.sort(subgraph_exec_times)

        self._total_exec_time = median_exec_time

        if local:
            if self.is_source:
                parents = []
            elif isinstance(self.parent, (list,tuple)):
                parents = self.parent
            else:
                parents = [self.parent,]

            parents_exec_time = sum([par.total_exec_time for par in parents if par.input_nodes])

            # parents = [par for par in self.all_parents.values() if par is not self]
            # parents_exec_time = sum([par.local_exec_time for par in parents if par.input_nodes])

            median_exec_time -= parents_exec_time
            self._local_exec_time = median_exec_time

        if nonegative and median_exec_time < 0:
            logger.warning('{}: Local execution time smaller than 0. "Correcting" to 0.'.format(self.name))
            median_exec_time = 0

        # Explicitly free memory since GPU memory tends to overflow here after multiple calls
        self._output_func.func = None
        for _ in range(3):  # Use this hack until https://github.com/Theano/Theano/issues/4275 is resolved.
            gc.collect()

        if print_info:
            logger.info('{0} samples in ms:\n{1}\n{0}: median execution time: {2} ms\n'
                        .format(self.name, subgraph_exec_times, median_exec_time))

        return median_exec_time

###############################################################################
###############################################################################
###############################################################################

class Input(Node):
    """
    Input Node

    Parameters
    ----------
    shape: list/tuple of int
        shape of input array, unspecified shapes are ``None``
    tags: list/tuple of strings or comma-separated string
        tags indicate which purpose the dimensions of the tensor serve. They are
        sometimes used to decide about reshapes. The maximal tensor has tags:
        "r, b, f, z, y, x, s" which denote:
        * r: perform recurrence along this axis
        * b: batch size
        * f: features, filters, channels
        * z: convolution no. 3 (slower than 1,2)
        * y: convolution no. 1
        * x: convolution no. 2
        * s: samples of the same instance (over which expectations are calculated)
        Unused axes are to be removed from this list, but ``b`` and ``f`` must
        always remain.
        To avoid bad memory layout, the order must not be changed.
        For less than 3 convolutions conv1,conv2 are preferred for performance reasons.
        Note that CNNs can mix nodes with 2d and 3d convolutions as 2d is a special
        case of 3d with filter size 1 on the respective axis. In this case conv3
        should be used for the axis with smallest filter size.
    strides
    fov
    dtype: str
        corresponding to numpy dtype (e.g., 'int64').
        Default is floatX from theano config
    hardcoded_shape
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, shape, tags, strides=None, fov=None, dtype=theano.config.floatX,
                 hardcoded_shape=False, name='input', print_repr=True):
        super(Input, self).__init__(None, name, print_repr)

        self.is_source = True
        self._shape = graphutils.TaggedShape(shape, tags, strides, fov=fov)
        if type(dtype) not in [str, unicode]:
            raise ValueError("dtype must be a string.")
        self.dtype = dtype
        self.hardcoded_shape = hardcoded_shape
        self._local_exec_time = 0


    def _make_output(self):
        broadcast = (False,)*len(self._shape)
        self.output = T.TensorType(self.dtype, broadcast)(name=self.name)
        if self.hardcoded_shape and not None in self._shape.shape:
            self.output = self.output.reshape(self._shape.shape)

    def _calc_shape(self):
        self.shape = self._shape

    def _calc_comp_cost(self):
        self.computational_cost = 0


def Input_like(ref, dtype=None, name='input',
               print_repr=True, override_f=False, hardcoded_shape=False):
    if isinstance(ref, Node):
        shape = list(ref.shape.shape)
        tags  = ref.shape.tags
        strides = ref.shape.strides
        fov = ref.shape.fov
        if override_f:
            shape = list(shape)
            shape[ref.shape.tag2index('f')] = override_f

        if dtype is None:
            dtype = ref.output.dtype

    elif isinstance(ref, graphutils.TaggedShape):
        shape = ref.shape
        tags  = ref.tags
        strides = ref.strides
        fov = ref.fov
        assert dtype is not None
        assert override_f is None
    else:
        raise ValueError("ref must be Node or TaggedShape.")
    return Input(shape, tags, strides, fov=fov, dtype=dtype,
                 name=name, print_repr=print_repr,
                 hardcoded_shape=hardcoded_shape)


class GenericInput(Node):
    """
    Input Node for arbitrary oject.

    Parameters
    ----------
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, name='generic_input',
                 print_repr=True):
        super(GenericInput, self).__init__(None, name, print_repr)
        self.is_source = True
        self._shape = None

    def _make_output(self):
        self.output = gof.type.Generic()(name=self.name)

    def _calc_shape(self):
        self.shape = self._shape


    def _calc_comp_cost(self):
        self.computational_cost = 0


class FromTensor(Node):
    """
    Dummy Node to be used in the split-function.

    Parameters
    ----------
    tensor: T.Tensor
    tensor_shape
    tensor_parent: T.Tensor
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, tensor, tensor_shape, tensor_parent, name='from_tensor',
                 print_repr=False):
        super(FromTensor, self).__init__(tensor_parent, name, print_repr)

        self._shape = tensor_shape
        self._output = tensor

    def _make_output(self):
        self.output = self._output

    def _calc_shape(self):
        self.shape = self._shape

    def _calc_comp_cost(self):
        self.computational_cost = 0


def split(node, axis='f', index=None, n_out=None, strip_singleton_dims=False, name='split'):
    args = ()
    kwargs = dict(axis=axis, index=index, n_out=n_out,
                  strip_singleton_dims=strip_singleton_dims, name=name)
    if isinstance(name, (list, tuple)):
        name_list = list(name)
        name = 'split'
    else:
        name_list = None
    model_manager.current.register_split(node, split, name, args, kwargs)

    sh = node.shape
    try:
        sh[axis]
    except:
        raise ValueError("%s has no index %s."%(sh, axis))

    if index is not None:
        index = utils.as_list(index)

    if (index is not None) and (np.max(index) >= sh[axis]):
        if sh[axis] is not None:
            raise ValueError("Index %s larger than size of axis %s (%s)."%(sh, axis, sh[axis]))

    if (n_out is not None) and (sh[axis]%n_out!=0):
        raise ValueError("Cannot split axis of size %i into %i parts."%(sh[axis], n_out))

    if isinstance(axis, (str, unicode)):
        axis = sh.tag2index(axis)

    if index is None:
        splits = [sh[axis]//n_out * i for i in range(n_out+1)]
    else:
        splits = [0,] + index + [sh[axis],]

    tensor = node.output
    out_nodes = []
    for i in range(len(splits)-1):
        if splits[i+1] is None:
            length = None
        else:
            length = splits[i + 1] - splits[i]

        sl = [slice(None),] * tensor.ndim
        if strip_singleton_dims and length==1:
            sl[axis] = splits[i]
        else:
            sl[axis] = slice(splits[i],splits[i+1])
        sub_tensor = tensor[tuple(sl)]

        if strip_singleton_dims and length==1:
            sub_sh = sh.delaxis(axis)
        else:
            sub_sh = sh.updateshape(axis, length)

        if name_list is not None:
            assert len(name_list)==len(splits)-1
            name_i = name_list[i]
        else:
            name_i = name
        out_node = FromTensor(sub_tensor, sub_sh, node, name=name_i)
        out_nodes.append(out_node)

    return tuple(out_nodes)


###############################################################################

class Concat(Node):
    """
    Node to concatenate the inputs. The inputs must have the same shape,
    except in the dimension corresponding to ``axis``. This is not checked as
    shapes might be unspecified prior to compilation!

    Parameters
    ----------
    parent_nodes: list of Node
        Inputs to be concatenated.
    axis: int
        Join axis.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, parent_nodes, axis='f', name="concat", print_repr=True):
        super(Concat, self).__init__(parent_nodes, name, print_repr)

        if not isinstance(parent_nodes, (tuple, list)):
            raise ValueError("Can only join list/tuple of nodes")

        if isinstance(axis, str):
            self.axis = parent_nodes[0].shape.tag2index(axis)
        else:
            self.axis = axis


    def _make_output(self):
        # It is assumed that all other dimensions are matching
        inputs = [inp.output for inp in self.parent]
        self.output = T.concatenate(inputs, axis=self.axis)

    def _calc_shape(self):
        joint_axis_size = reduce(lambda x,y: x+y.shape[self.axis],
                                 self.parent, 0)
        # assuming all other dimensions are equal
        sh = self.parent[0].shape.updateshape(self.axis, joint_axis_size)
        self.shape = sh

    def _calc_comp_cost(self):
        self.computational_cost = 0

    ###############################################################################

class MultMerge(Node):
    """
    Node to concatenate the inputs. The inputs must have the same shape,
    except in the dimension corresponding to ``axis``. This is not checked as
    shapes might be unspecified prior to compilation!

    Parameters
    ----------
    n1: Node
        First input node.
    n2: Node
        Second input node.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, n1, n2, name="multmerge",
                 print_repr=True):
        super(MultMerge, self).__init__((n1, n2), name, print_repr)
        assert n1.shape.shape==n2.shape.shape

    def _make_output(self):
        self.output = T.mul(self.parent[0].output, self.parent[1].output)

    def _calc_shape(self):
        self.shape = self.parent[0].shape.copy()

    def _calc_comp_cost(self):
        self.computational_cost = self.parent[0].shape.stripnone_prod

###############################################################################

class ApplyFunc(Node):
    """
    Apply function to the input. If the function changes the output shape,
    this node should not be used.

    Parameters
    ----------
    parent: Node
        Input (single).
    functor: function
        Function that acts on theano variables (e.g. ``theano.tensor.tanh``).
    args: tuple
        Arguments passed to ``functor`` **after** the input.
    kwargs: dict
        kwargs for ``functor``.
    """

    def __init__(self, parent, functor, args=(), kwargs={}, name="apply",
                 print_repr=False):
        super(ApplyFunc, self).__init__(parent, name, print_repr)

        self._functor = functor
        self._args = args
        self._kwargs = kwargs


    def _make_output(self):
        self.output = self._functor(self.parent.output,
                                    *self._args, **self._kwargs)


class ValueNode(Node):
    """
    (Optinally) trainable Value Node

    Parameters
    ----------
    shape: list/tuple of int
        shape of input array, unspecified shapes are ``None``
    tags: list/tuple of strings or comma-separated string
        tags indicate which purpose the dimensions of the tensor serve. They are
        sometimes used to decide about reshapes. The maximal tensor has tags:
        "r, b, f, z, y, x, s" which denote:
            * r: perform recurrence along this axis
            * b: batch size
            * f: features, filters, channels
            * z: convolution no. 3 (slower than 1,2)
            * y: convolution no. 1
            * x: convolution no. 2
            * s: samples of the same instance (over which expectations are calculated)
        Unused axes are to be removed from this list, but ``b`` and ``f`` must
        always remain.
        To avoid bad memory layout, the order must not be changed.
        For less than 3 convolutions conv1,conv2 are preferred for performance reasons.
        Note that CNNs can mix nodes with 2d and 3d convolutions as 2d is a special
        case of 3d with filter size 1 on the respective axis. In this case conv3
        should be used for the axis with smallest filter size.
    strides
    fov
    dtype: str
        corresponding to numpy dtype (e.g., 'int64').
        Default is floatX from theano config
    apply_train: bool
    value
    init_kwargs: dict
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, shape, tags, strides=None, fov=None, dtype=theano.config.floatX, apply_train=False,
                 value=None, init_kwargs=None, name='value_state', print_repr=True):
        super(ValueNode, self).__init__(None, name, print_repr)

        if isinstance(value, list): # Handle locking of params
            if value[1]=='const' and len(value==2):
                value = value[0]
                apply_train = False

        self.is_source = False
        self._shape = graphutils.TaggedShape(shape, tags, strides, fov=fov)
        if type(dtype)!=str:
            raise ValueError("dtype must be a string.")
        self._dtype = dtype
        self._apply_train = apply_train
        self._value = value
        self._init_kwargs = init_kwargs

    def _make_output(self):
        broadcast = (False,)*len(self._shape)
        p_name = '<%s%s>' % ('state_', tuple(self._shape))
        if self._apply_train:
            self.output = VariableWeight(self._shape, dtype=self._dtype, value=self._value,
                                         init_kwargs=self._init_kwargs, name=p_name, broadcastable=broadcast)
        else:
            if self._init_kwargs is not None:
                raise NotImplementedError()

            if self._value is None:
                self._value = np.zeros(self._shape)
            else:
                if not isinstance(self._value, np.ndarray):
                    self._value = np.array(self._value)

            self.output = ConstantParam(self._value, name=p_name,
                                        dtype=self._dtype,
                                        make_singletons_broadcastable=False)
        self.params['value'] = self.output

    def _calc_shape(self):
        self.shape = self._shape

    def _calc_comp_cost(self):
        self.computational_cost = 0

    def get_value(self):
        return self.output.get_value()


class InitialState_like(Node):
    """

    Parameters
    ----------
    parent
    override_f
    dtype
    name
    print_repr
    init_kwargs
    """

    def __init__(self, parent, override_f, dtype=None, name='initial_state',
               print_repr=True, init_kwargs=None):
        assert isinstance(parent, Node)
        super(InitialState_like, self).__init__(parent, name, print_repr)

        if init_kwargs is None:
            init_kwargs = dict(mode='const', scale=0.0)

        self._init_kwargs = init_kwargs
        shape = list(parent.shape.shape)
        tags  = parent.shape.tags
        strides = parent.shape.strides
        fov = parent.shape.fov
        if dtype is None:
            dtype = parent.output.dtype

        self.axis = parent.shape.tag2index('f')

        shape[self.axis] = override_f
        self._shape = graphutils.TaggedShape(shape, tags, strides, fov=fov)
        self._dtype = dtype
        self.override_f = override_f

    def _make_output(self):
        p_name = '<%s%s>' % ('initial_state_', self.override_f)
        single = VariableWeight(shape=self.override_f, dtype=self._dtype,
                                     init_kwargs=self._init_kwargs,
                                     name=p_name)
        self.params['value'] = single
        pattern = ['x' for i in self.parent.shape]
        pattern[self.axis] = 0
        single = single.dimshuffle(pattern)
        sh = list(self.parent.output.shape)
        sh[self.axis] = 1
        broadcaster = T.ones(sh, dtype=self._dtype)
        self.output = single * broadcaster

    def _calc_shape(self):
        self.shape = self._shape

    def _calc_comp_cost(self):
        self.computational_cost = 0

    def get_value(self):
        return self.params['value'].get_value()





# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

from collections import OrderedDict
import sys
import time
import logging

import numpy as np
import theano

logger = logging.getLogger('elektronn2log')

if sys.version_info.major >= 3:
    unicode = str

__all__ = ['TaggedShape', 'floatX', 'as_floatX', 'make_func',
           'getinput_for_multioutput']

floatX = theano.config.floatX

class TaggedShape(object):
    """
    Object to manage shape and associated tags uniformly.
    The ``[]``-operator can be used get shape values by either index (``int``) or tag (``str``)

    Parameters
    ----------
    shape: list/tuple of int
        shape of array, unspecified shapes are ``None``
    tags: list/tuple of strings or comma-separated string
        tags indicate which purpose the dimensions of the tensor serve. They are
        sometimes used to decide about reshapes. The maximal tensor has tags:
        "r, b, s, f, z, y, x, s" which denote:
        * r: perform recurrence along this axis
        * b: batch size
        * s: samples of the same instance (over which expectations are calculated)
        * f: features, filters, channels
        * z: convolution no. 3 (slower than 1,2)
        * y: convolution no. 1
        * x: convolution no. 2
    strides:
        list of strides, only for spatial dimensions, so it is 1-3 long
    mfp_offsets:
    """

    def __init__(self, shape, tags, strides=None, mfp_offsets=None, fov=None,):
        self._shape = list(shape) # copy
        self._tags  = self._check_tags(tags) # is copied too
        if len(self._shape)!=len(self._tags):
            raise ValueError("Shape %s and tags %s must have same length"\
                             %(self._shape, self._tags))

        if strides is None:
            self._strides = np.ones(len(self.spatial_axes), np.int)
        else:
            self._strides = np.array(strides, np.int)

        if mfp_offsets is None:
            self._mfp_offsets = np.zeros((1, len(self.spatial_axes)), np.int)
        else:
            self._mfp_offsets = np.atleast_2d((np.array(mfp_offsets, np.int)))

        if fov is None:
            self._fov = np.ones(len(self.spatial_axes), np.int)
        else:
            self._fov = np.array(fov, np.int)

    def __repr__(self):
        r = "["
        for s,t  in zip(self._shape, self._tags):
            r += "(%s,%s), "%(s,t)

        r = r[:-2] # remove last 2 chars
        r += "]"
        return r

    @property
    def ext_repr(self):
        s = repr(self)
        s += '\n'
        s += 'fov=%s, offsets=%s, strides=%s, spatial shape=%s'\
             %(self.fov, self.offsets, self.strides, self.spatial_shape)
        return s


    @staticmethod
    def _check_tags(tags):
        if tags is None:
            return None

        if not isinstance(tags, (list, tuple)):
            if isinstance(tags, (str, unicode)):
                tags = tags.split(',')
                tags = [x.strip() for x in tags]
            else:
                raise ValueError("Tags must be either list/tuple of comma-separated string, not %s" %tags)

        allowed_tags = ['r', 'b', 'z', 'f', 'x', 'y', 's']
        for t in tags:
            if t not in allowed_tags:
                raise ValueError("Unknown tag %s" %(t,))

        return tags

    def __getitem__(self, slice):
        if isinstance(slice, (str, unicode)):
            return self._shape[self.tag2index(slice)]
        else:
            return self._shape[slice]

    def __len__(self):
        return len(self._shape)

    @property
    def shape(self):
        return self._shape

    @property
    def tags(self):
        return self._tags

    @property
    def strides(self):
        return self._strides

    @property
    def mfp_offsets(self):
        return self._mfp_offsets

    @property
    def fov(self):
        return list(self._fov)

    @property
    def fov_all_centered(self):
        fov = self.fov
        return np.all(np.mod(fov, 2)==1)

    @property
    def offsets(self):
        return [i//2 for i in self.fov]

    @property
    def spatial_axes(self):
        spatial_axes = [tag for tag in ['z', 'x', 'y'] if self.hastag(tag)]
        spatial_axes = [self.tag2index(tag) for tag in spatial_axes]
        spatial_axes.sort()
        return list(spatial_axes)

    @property
    def ndim(self):
        return len(self.spatial_axes)

    @property
    def spatial_shape(self):
        return [self.shape[i] for i in self.spatial_axes]

    @property
    def spatial_size(self):
        return int(np.prod(self.spatial_shape))

    @property
    def stripnone(self):
        """
        Return the shape but with all None elements removed (e.g. if batch size
        is unspecified)
        """
        new_sh = []
        for s in self._shape:
            if s is not None:
                new_sh.append(s)

        return new_sh

    @property
    def stripbatch_prod(self):
        """
        Calculate product excluding batch dimension
        """
        new_sh = []
        for s,t in zip(self._shape, self._tags):
            if t is not 'b':
                new_sh.append(s)

        return np.prod(new_sh)

    @property
    def stripnone_prod(self):
        """
        Return the product of the shape but with all None elements removed
        (e.g. if batch size is unspecified)
        """
        return np.prod(self.stripnone)

    def tag2index(self, target_tag):
        """
        Finds the index of the desired tag
        """
        try:
            i = self._tags.index(target_tag)
        except ValueError:
            raise ValueError("Shape does not have tag %s, only"
                             "tags %s" %(target_tag, self._tags))

        return i

    def hastag(self, tag):
        has = False
        try:
            self.tag2index(tag)
            has = True
        except ValueError:
            pass

        return has


    def updateshape(self, axis, new_size, mode=None):
        """
        Create new TaggedShape with ``new_size`` on ``axis`` .
        Modes for updating: ``None`` (override), ``'add'``, ``'mult'``
        """
        if isinstance(axis, int):
            i = axis
        else:
            i = self.tag2index(axis)

        sh = list(self._shape) # copy

        if mode is None:
            sh[i] = new_size
        else:
            if sh[i] is None:
                logger.debug("Updating shape %s at index %i carrying value None, "
                             "will stay None" %(self, i))
                sh[i] = sh[i]
            elif mode=='add':
                sh[i] += new_size
            elif mode=='mult':
                sh[i] *= new_size

        ret = self.copy()
        ret._shape = sh

        return ret


    def updatefov(self, axis, new_fov):
        """
        Create new TaggedShape with ``new_fov`` on ``axis``. Axis is given as
        index of the spatial axes (not matching the absolute index of sh).
        """
        ret = self.copy()
        ret._fov[axis] = new_fov

        return ret

    def updatestrides(self, strides):
        ret = self.copy()
        ret._strides = strides
        return ret

    def updatemfp_offsets(self, mfp_offsets):
        ret = self.copy()
        ret._mfp_offsets = mfp_offsets
        return ret

    def addaxis(self, axis, size, tag):
        """
        Create new TaggedShape with new axis inserted at ``axis`` of size ``size``
        tagged ``tag``. If axis is a tag, the new axis is **right** of that tag
        """
        if isinstance(axis, int):
            i = axis
        else:
            i = self.tag2index(axis) + 1

        sh = list(self._shape) # copy
        sh.insert(i, size)
        tags = list(self._tags)
        tags.insert(i, tag)

        return TaggedShape(sh, tags, self._strides, self._mfp_offsets, self._fov)

    def delaxis(self, axis):
        """
        Create new TaggedShape with new axis inserted at ``axis`` of size ``size``
        tagged ``tag``. If axis is a tag, the new axis is **right** of that tag
        """
        if isinstance(axis, int):
            i = axis
        else:
            i = self.tag2index(axis) + 1

        sh = list(self._shape) # copy
        sh.pop(i)
        tags = list(self._tags)
        tags.pop(i)
        return TaggedShape(sh, tags, self._strides, self._mfp_offsets, self._fov)


    def copy(self):
        return TaggedShape(self._shape, self._tags,
                           self._strides, self._mfp_offsets, self.fov)



class make_func(object):
    """
    Wrapper for compiled theano functions. Features:

    * The function is compiled on demand (i.e. no wait at initialisation)
    * Singleton return values are returned directly, multiple values as list
    * The last execution time can inspected in the attribute ``last_exec_time``
    * Functions can be timed: ``profile_execution`` is an ``int`` that specifies
      the number of runs to average. The average time is printed then.
    * In/Out values can have a ``borrow`` flag which might overwrite the numpy
      arrays but might speed up execution (see theano doc)
    """

    def __init__(self, tt_input, tt_output, updates=None, name='Unnamed Function',
                 borrow_inp=False, borrow_out=False, profile_execution=False):
        self.name = name
        self.func = None
        self.profile = profile_execution
        self.last_exec_time = None
        self.updates = updates
        if borrow_inp:
            tt_input = [theano.In(x, borrow=True) for x in tt_input]

        self.tt_input = tt_input

        self.single_return = False
        if not isinstance(tt_output, (list, tuple)):
            tt_output = [tt_output,]
            self.single_return = True

        if borrow_out:
            tt_output = [theano.Out(x, borrow=True) for x in tt_output]

        self.tt_output = tt_output

    def __call__(self, *args):
        if self.func is None:
            self.compile()

        if len(args)==0: # calling foo() now just compiles the function
            return

        if self.profile:
            t0 = time.time()
            if self.profile>1:
                for i in range(self.profile-1):
                    self.func(*args)

            ret = self.func(*args)
            t = time.time() - t0
            logger.info("Function <%s> took %.5f s averaged over %i execs" % (self.name, t/self.profile, self.profile))

        else:
            t0 = time.time()
            ret =  self.func(*args)
            t = time.time() - t0
            self.last_exec_time = t

        if self.single_return:
            ret = ret[0]
        return ret

    def compile(self, profile=False):
        t_init = time.time()
        logger.info("Compiling %s, inputs=%s" % (self.name, self.tt_input))
        tf = theano.function(self.tt_input,
                             self.tt_output,
                             updates=self.updates,
                             name=self.name,
                             on_unused_input='warn',
                             profile=profile)

        logger.info(" Compiling done  - in %.2f s" % (time.time() - t_init))
        self.func = tf


def getinput_for_multioutput(outputs):
    """
    For list of several output layers return a list of required
    input tensors (without duplicates) to compute these outputs.
    """
    inputs = [out.input_tensors for out in outputs]
    ret = OrderedDict()
    for inp in inputs:
        for i in inp:
            ret[i] = True

    return list(ret.keys())


def as_floatX(x):
    if not hasattr(x, '__len__'):
        return np.array(x, dtype=floatX)
    return np.ascontiguousarray(x, dtype=floatX)





# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

import logging
import numpy as np
import theano.tensor as T

from . import graphutils
from . import variables

logger = logging.getLogger('elektronn2log')


class Optimiser(object):
    global_lr = variables.VariableParam(value=1,
                                        name='lr',
                                        dtype=graphutils.floatX)
    global_weight_decay = variables.VariableParam(value=0,
                                                  name='weight_decay',
                                                  dtype=graphutils.floatX)

    global_mom = variables.VariableParam(value=0.9,
                                          name='mom',
                                          dtype=graphutils.floatX)

    @classmethod
    def setlr(cls, val):
        """
        Set learning rate (global to all optimisers)
        """
        val = graphutils.as_floatX(val)
        cls.global_lr.set_value(val)

    @classmethod
    def setwd(cls, val):
        """
        Set weight decay parameter (global to all optimisers)
        """
        val = graphutils.as_floatX(val)
        cls.global_weight_decay.set_value(val)

    @classmethod
    def setmom(cls, val):
        """
        Set momentum parameter (global to all optimisers)
        """
        val = graphutils.as_floatX(val)
        cls.global_mom.set_value(val)

    def __init__(self, inputs, loss, grads, params, additional_outputs):
        if additional_outputs is None:
            additional_outputs = []

        self.meta_params = dict(lr=self.global_lr,
                                mom=self.global_mom,
                                wd=self.global_weight_decay)
        self.input = inputs
        self.output = [loss,] + additional_outputs
        self.loss   = loss
        self.params = params
        self.grads = grads
        self.step = None
        self.last_exec_time = None
        self.last_dir = []
        # the higher the index the older the params
        self.params_cycler = [self.alloc_shared_grads(name_suffix='_lp_%i'%i)
                              for i in range(3)]

    def alloc_shared_grads(self, name_suffix='_lg', init_val=0.0):
        """Returns new shared variables matching the shape of params/gradients"""
        grads = []
        for i, p in enumerate(self.params):
            name = p.name+name_suffix
            value = np.ones_like(p.get_value()) * graphutils.as_floatX(init_val)
            g = variables.VariableParam(value=value, name=name)
            grads.append(g)

        return grads

    def set_opt_meta_params(self, value_dict):
        """
        Update the meta-parameters via value dictionary
        """
        for k,v in value_dict.items():
            try:
                self.meta_params[k].set_value(v)
            except AttributeError:
                raise AttributeError

    def clear_last_dir(self, last_dir=None):
        if last_dir is None:
            last_dir = self.last_dir

        for d in last_dir:
            d.set_value(np.zeros(d.get_value().shape, dtype=d.dtype))


    def get_rotational_updates(self):
        updates = []
        for x in zip(self.params, *self.params_cycler):
            new_param, param_queue, = x[0], x[1:]
            for i in range(len(param_queue)-1, 0, -1):
                updates.append((param_queue[i], param_queue[i-1]))

            updates.append((param_queue[0], new_param))

        return updates


    def repair(self):
        self.clear_last_dir()
        for p, p_old in zip(self.params, self.params_cycler[-1]):
            p.set_value(p_old.get_value())


    def __call__(self, *args):
        """
        Perform an update step
        [data (,labels etc...)] --> [loss (, add. outputs...)]
        """
        ret = list(self.step(*args))
        ret[0] = graphutils.as_floatX(ret[0]) # the scalar loss
        self.last_exec_time = self.step.last_exec_time
        return ret



###############################################################################

class SGD(Optimiser):
    """
    SGD optimiser (See https://en.wikipedia.org/wiki/Stochastic_gradient_descent).
    """
    def __init__(self,  inputs, loss, grads, params, extra_updates,
                 additional_outputs=None):
        super(SGD, self).__init__(inputs, loss, grads, params,
                                  additional_outputs)
        self.last_dir     = self.alloc_shared_grads() # last direction os update

        updates = []
        for g, d, p in zip(self.grads, self.last_dir, self.params):
            new_d = g + self.global_mom * d
            if p.apply_reg:
                if p.apply_reg > 1:
                    multiplier = graphutils.as_floatX(p.apply_reg)
                    new_p = p - self.global_lr * \
                                (new_d + self.global_weight_decay * p * multiplier)
                else:
                    new_p = p - self.global_lr * \
                                (new_d + self.global_weight_decay * p)
            else:
                new_p = p - self.global_lr *  new_d

            updates.append((d, new_d))
            updates.append((p, new_p))

        updates.extend(extra_updates)
        updates.extend(self.get_rotational_updates())
        self.step = graphutils.make_func(self.input, self.output,
                                 updates=updates, name='SGD step')


class AdaGrad(Optimiser):
    """
    AdaGrad optimiser (See http://jmlr.org/papers/v12/duchi11a.html).

    Tries to favor making faster progress on parameters with usually small
    gradients (but does somehow ignore their actual direction, i.e. a parameter
    which has a lot of small gradients in the same direction and one that has
    many small gradients in opposite directions have both a high LR !
    """
    def __init__(self, inputs, loss, grads, params, extra_updates,
                 additional_outputs=None):
        super(AdaGrad, self).__init__(inputs, loss, grads, params,
                                      additional_outputs)

        self._init_done = False
        self.hs  = self.alloc_shared_grads('_h', init_val=0.0)
        updates = []
        for g, h, p in zip(self.grads, self.hs, self.params):
            new_h = h + T.square(g)
            if p.apply_reg: # apply to W but not b
                new_p = p - self.global_lr / T.sqrt(new_h) * \
                            (g + self.global_weight_decay * p)
            else:
                new_p = p - self.global_lr / T.sqrt(new_h) * g

            updates.append((h, new_h))
            updates.append((p, new_p))

        updates.extend(extra_updates)
        self.step = graphutils.make_func(self.input, self.output,
                                 updates=updates, name='AdaGrad step')

        # Create init_func to init h from one gradient evaluation
        updates = []
        for g, h in zip(self.grads, self.hs):
            new_h = h + T.square(g)
            updates.append((h, new_h))

        self.init_func = graphutils.make_func(self.input, [], updates=updates,
                                     name='AdaGrad initialiser')

    def __call__(self, *args):
        if not self._init_done:
            self.init_func(*args)
            self._init_done = True

        return super(AdaGrad, self).__call__(*args)


    def repair(self):
        super(AdaGrad, self).repair()
        self.clear_last_dir(self.hs)
        self._init_done = False

class AdaDelta(Optimiser):
    """
    AdaDelta optimiser (See https://arxiv.org/abs/1212.5701).

    Like AdaGrad, but accumulate squared only over windo
    The delta part is some diagonal hessian approximation.
    Claims to be robust agains sudden large gradients because then the
    denominator explodes, but this explosion is persistent for a while...
    (and this argumentation is true for any method accumulating squared grads).
    """
    def __init__(self,  inputs, loss, grads, params, extra_updates,
                 additional_outputs=None):
        super(AdaDelta, self).__init__(inputs, loss, grads, params,
                                       additional_outputs)
        self.squared_accum = self.alloc_shared_grads("_sq") # last directions update
        self.delta_accum = self.alloc_shared_grads("_d") # last directions update
        epsilon = 1e-5

        updates = []
        for g, s, d, p in zip(self.grads, self.squared_accum,
                              self.delta_accum, self.params):
            new_s = self.global_mom * s + (1.0 - self.global_mom) * T.square(g)
            direction = (g * T.sqrt(d + epsilon) / T.sqrt(s + epsilon))
            new_d = self.global_mom * d + (1 - self.global_mom) * T.square(direction)
            if p.apply_reg:
                if p.apply_reg > 1:
                    multiplier = graphutils.as_floatX(p.apply_reg)
                    new_p = p - self.global_lr * \
                                (direction + self.global_weight_decay * p * multiplier)
                else:
                    new_p = p - self.global_lr * \
                                (direction + self.global_weight_decay * p)
            else:
                new_p = p - self.global_lr *  direction

            updates.append((s, new_s))
            updates.append((d, new_d))
            updates.append((p, new_p))

        updates.extend(extra_updates)
        updates.extend(self.get_rotational_updates())
        self.step = graphutils.make_func(self.input, self.output,
                                 updates=updates, name='AdaDelta step')


    def repair(self):
        super(AdaDelta, self).repair()
        self.clear_last_dir(self.squared_accum)
        self.clear_last_dir(self.delta_accum)


class Adam(Optimiser):
    """
    Adam optimiser (See https://arxiv.org/abs/1412.6980v9).

    Like AdaGrad with windowed squared_accum and with momentum and a bias for
    the initial phase (t).
    The normalisation of Adam and AdaGrad (and RMSProp) does not damp but
    exaggerates sudden steep gradients (their squared_accum is small and their
    current grad is large).
    """
    def __init__(self,  inputs, loss, grads, params, extra_updates,
                 additional_outputs=None):
        super(Adam, self).__init__(inputs, loss, grads, params,
                                   additional_outputs)
        self.squared_accum = self.alloc_shared_grads("_sq") # last directions update
        self.momentum = self.alloc_shared_grads("_m") # last directions update
        epsilon = 1e-5

        # self.beta1 = variables.VariableParam(value=0.9, name='beta1',
        #                                      dtype=graphutils.floatX)
        self.beta2 = variables.VariableParam(value=0.999, name='beta2',
                                             dtype=graphutils.floatX)
        #self.meta_params['beta1'] = self.beta1
        self.meta_params['beta2'] = self.beta2

        t_old = variables.VariableParam(value=0.0, name='beta2',
                                        dtype=graphutils.floatX)

        updates = []
        t = 1 + t_old
        updates.append((t_old, t))
        factor = T.sqrt(1-self.beta2**t)/(1-self.global_mom**t)
        for g, s, m, p in zip(self.grads, self.squared_accum,
                              self.momentum, self.params):
            new_m = self.global_mom * m + (1.0 - self.global_mom) * g
            new_s = self.beta2      * s + (1.0 - self.beta2) * T.square(g)

            direction = factor * new_m / T.sqrt(new_s + epsilon)
            if p.apply_reg:
                if p.apply_reg > 1:
                    multiplier = graphutils.as_floatX(p.apply_reg)
                    new_p = p - self.global_lr * \
                                (direction + self.global_weight_decay * p * multiplier)
                else:
                    new_p = p - self.global_lr * \
                                (direction + self.global_weight_decay * p)
            else:
                new_p = p - self.global_lr *  direction

            updates.append((s, new_s))
            updates.append((m, new_m))
            updates.append((p, new_p))

        updates.extend(extra_updates)
        updates.extend(self.get_rotational_updates())
        self.step = graphutils.make_func(self.input, self.output,
                                 updates=updates, name='Adam step')

    def repair(self):
        super(Adam, self).repair()
        self.clear_last_dir(self.squared_accum)
        self.clear_last_dir(self.momentum)

class CG(Optimiser):
    pass
